# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import base64
import requests.models
from datetime import timedelta

import contrast
from contrast.agent.disable_reaction import DisableReaction
from contrast.agent.settings import Settings
from contrast.utils.object_utils import NOTIMPLEMENTED_MSG
from contrast.utils.decorators import fail_loudly
from contrast.utils.timer import now_ms, sleep

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


PYTHON = "Python"
SLEEP_TIME_SECS = timedelta(minutes=15).seconds


def _b64url_stripped(header_str):
    """
    For some headers, TS expects a value that
    - is base64 encoded using URL-safe characters
    - has any padding (= or ==) stripped

    This follows RFC-4648 - base64 with URL and filename safe alphabet
    """
    return base64.urlsafe_b64encode(header_str.encode()).rstrip(b"=").decode("utf-8")


class BaseTsMessage:
    def __init__(self):
        self._sent_count = 0
        self.settings = Settings()

        self.base_url = f"{self.settings.api_url}/api/ng/"
        self.proxy = (
            self.settings.build_proxy_url() if self.settings.is_proxy_enabled else {}
        )

        self.server_name_b64 = _b64url_stripped(self.settings.get_server_name())
        self.server_path_b64 = _b64url_stripped(self.settings.get_server_path())
        self.server_type_b64 = _b64url_stripped(self.settings.get_server_type())
        auth_header = f"{self.settings.api_user_name}:{self.settings.api_service_key}"

        self.headers = {
            # the Authorization header must not have its padding stripped
            "Authorization": base64.urlsafe_b64encode(auth_header.encode()).decode(),
            "API-Key": self.settings.api_key,
            "Server-Name": self.server_name_b64,
            "Server-Path": self.server_path_b64,
            "Server-Type": self.server_type_b64,
            "X-Contrast-Agent": f"{PYTHON} {contrast.__version__}",
            "X-Contrast-Header-Encoding": "base64",
        }

        self.body = ""

    @property
    def class_name(self):
        return type(self).__name__.lstrip("_")

    @property
    def name(self) -> str:
        """
        Used for request audit filename
        """
        raise NotImplementedError(NOTIMPLEMENTED_MSG)

    @property
    def path(self) -> str:
        """
        URL path for teamserver; used for formatting as "/api/ng/{path}"
        """
        raise NotImplementedError(NOTIMPLEMENTED_MSG)

    @property
    def request_method(self) -> str:
        raise NotImplementedError(NOTIMPLEMENTED_MSG)

    @property
    def expected_response_codes(self):
        return [204]

    @property
    def disable_agent_on_401_and_408(self):
        return False

    @property
    def sent_count(self):
        return self._sent_count

    def sent(self):
        self._sent_count += 1

    @fail_loudly("Failed to process TS response")
    def process_response(
        self, response: requests.models.Response, reporting_client
    ) -> None:
        del response, reporting_client
        raise NotImplementedError(NOTIMPLEMENTED_MSG)

    def should_shutdown(self, response: requests.models.Response) -> bool:
        """
        Validate 404 NotFoundApplication or NotFoundServer response

        From the spec:

        "Note that the agent should only consider this a valid response if the body
        is present and matches the schema defined for `ResponseMessage`.  Specifically,
        an agent should parse this response and check for a `success` value of `false`
        before disabling."

        https://github.com/Contrast-Security-Inc/contrast-agent-api-spec/blob/master/agent-endpoints.yml
        """
        try:
            body = response.json()
            return body.get("success") == False
        except Exception:
            pass

        return False

    def process_response_code(
        self, response, reporting_client
    ):  # pylint: disable=too-many-return-statements
        """
        Return True if response code is expected response code
        """
        if not isinstance(response, requests.models.Response):
            return False

        logger.debug(
            "%s: received %s response code from Teamserver",
            self.class_name,
            response.status_code,
        )

        if response.status_code in (204, 304):
            # Both of these codes indicate no content, meaning there is no
            # action for us to take. Nothing has changed on TeamServer for
            # us to process.
            return False

        if response.status_code == 404:
            if self.should_shutdown(response):
                DisableReaction.run(self.settings)
            return False

        if response.status_code in (409, 410, 412, 502):
            # 409: app is archived, 502 app is locked in TS
            # 410: app is not registered. We could send App startup for not we won't
            # 412: API key no longer valid. While spec may say to resend msg in 15 mins,
            #  in reality the app server and agent should simply be restarted.
            DisableReaction.run(self.settings)
            return False

        if response.status_code in (401, 408):
            # 401: Access forbidden because credentials failed to authenticate.
            # 408: TS Could not create settings in time.
            if self.disable_agent_on_401_and_408:
                DisableReaction.run(self.settings)
            else:
                logger.debug("Sleeping for 15 minutes")

                sleep(SLEEP_TIME_SECS)

                reporting_client.retry_message(self)

            return False

        if response.status_code == 429:
            sleep_time = int(response.headers.get("Retry-After", SLEEP_TIME_SECS))

            logger.debug("Sleeping for %s seconds", sleep_time)

            sleep(sleep_time)

            reporting_client.retry_message(self)

        if response.status_code not in self.expected_response_codes:
            logger.debug(
                "Unexpected %s response from TS: %s",
                self.class_name,
                response.status_code,
            )
            return False

        return True


class BaseTsServerMessage(BaseTsMessage):
    @fail_loudly(f"Failed to process server settings response")
    def process_response(self, response, reporting_client):
        settings = Settings()
        if not self.process_response_code(response, reporting_client):
            return

        body = response.json()

        settings.apply_ts_feature_settings(body)
        settings.process_ts_reactions(body)


class BaseTsAppMessage(BaseTsMessage):
    def __init__(self):
        super().__init__()

        # App language should only be encoded for url paths, not for headers.
        self.app_language_b64 = _b64url_stripped(PYTHON)
        self.app_name_b64 = _b64url_stripped(self.settings.app_name)

        self.headers.update(
            {
                "Application-Language": PYTHON,
                "Application-Name": self.app_name_b64,
                "Application-Path": _b64url_stripped(self.settings.app_path),
            }
        )

    @property
    def since_last_update(self):
        """
        Time in ms since app settings have been updated.
        If never updated, then it's been 0ms since then.
        """
        if self.settings.last_app_update_time_ms == 0:
            return 0
        return now_ms() - self.settings.last_app_update_time_ms



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import ctypes
import functools
from enum import IntEnum

from contrast.agent import agent_lib
from contrast.api import TypeCheckedProperty
from contrast.api.user_input import InputType
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")

# These are rules we do not have an implementation for yet
# Other rule IDs where added directly into the protect rule class
SSJS_INJECTION_RULE_ID = 1 << 7

# This should not be used directly. It's used to construct INPUT_TYPE_LOOKUP below.
_AGENT_LIB_TO_TS_INPUT_TYPE = {
    "CookieName": InputType.COOKIE_NAME,
    "CookieValue": InputType.COOKIE_VALUE,
    "HeaderKey": InputType.HEADER,
    "HeaderValue": InputType.HEADER,
    "JsonKey": InputType.JSON_VALUE,  # there is no such thing as InputType.JSON_KEY
    "JsonValue": InputType.JSON_VALUE,
    "Method": InputType.METHOD,
    "ParameterKey": InputType.PARAMETER_NAME,
    "ParameterValue": InputType.PARAMETER_VALUE,
    "UriPath": InputType.URI,
    "UrlParameter": InputType.URL_PARAMETER,
    "MultipartName": InputType.MULTIPART_NAME,
    "XmlValue": InputType.XML_VALUE,
}


class DBType(IntEnum):
    DB2 = 1
    MYSQL = 2
    ORACLE = 3
    POSTGRES = 4
    SQLITE = 5
    SQL_SERVER = 6
    UNKNOWN = 7

    @staticmethod
    def from_str(label):
        label = label.upper()
        try:
            return DBType[label]
        except KeyError:
            if label == "SQLITE3":
                return DBType.SQLITE
            if label == "POSTGRESQL":
                return DBType.POSTGRES
            if label in ("SQL SERVER", "SQL_SERVER", "SQLSERVER"):
                return DBType.SQL_SERVER

            return DBType.UNKNOWN


class CCheckQuerySinkResult(ctypes.Structure):
    _fields_ = [
        ("start_index", ctypes.c_ulonglong),
        ("end_index", ctypes.c_ulonglong),
        ("boundary_overrun_index", ctypes.c_ulonglong),
        ("input_boundary_index", ctypes.c_ulonglong),
    ]


@functools.lru_cache(maxsize=1)
def rule_id_lookup():
    """
    we'd like this to be a module-level dict, but we need to load this after agent_lib
    initialization
    """
    return {
        rule_int: rule_str
        for rule_str, rule_int in agent_lib.modules.CONSTANTS.RuleType.items()
    }


@functools.lru_cache(maxsize=1)
def input_type_lookup():
    """
    agent lib input type int -> InputType enum (TS representation)

    we'd like this to be a module-level dict, but we need to load this after agent_lib
    initialization
    """
    return {
        input_type_int: _AGENT_LIB_TO_TS_INPUT_TYPE.get(
            input_type_str, InputType.UNKNOWN
        )
        for input_type_str, input_type_int in agent_lib.modules.CONSTANTS.InputType.items()
    }


class InputAnalysisResult:
    rule_id = TypeCheckedProperty(str)
    input_type = TypeCheckedProperty(InputType, constructor_arg=InputType.UNKNOWN)
    score = TypeCheckedProperty(float)
    key = TypeCheckedProperty(str)
    value = TypeCheckedProperty(str)
    path = TypeCheckedProperty(str)
    ids = TypeCheckedProperty(list)
    attack_count = TypeCheckedProperty(int)
    prefer_worth_watching = TypeCheckedProperty(bool)

    def __init__(self, input_value, value, ceval_results, prefer_worth_watching):
        if (
            isinstance(ceval_results, agent_lib.modules.CONSTANTS.CEvalResult)
            and input_value is not None
            and isinstance(value, str)
        ):
            self.rule_id = rule_id_lookup()[ceval_results.rule_id]
            self.input_type = input_type_lookup()[ceval_results.input_type]

            self.score = ceval_results.score

            if isinstance(input_value, str):
                self.key = input_value
            elif isinstance(input_value, bytes):
                self.key = input_value.decode()
            else:
                # c_char_p
                self.key = input_value.value.decode()

            self.value = value
            self.prefer_worth_watching = prefer_worth_watching


class SQLInjectionResult:
    def __init__(self, sql_query, input_index, input_len, ccheck_query_sink_result):
        self.boundary_overrun_index = None
        self.end_index = None
        self.input_boundary_index = None
        self.start_index = None
        self.input_query = None
        self.input_index = None
        self.input_len = None

        if (
            isinstance(ccheck_query_sink_result, CCheckQuerySinkResult)
            and isinstance(sql_query, str)
            and isinstance(input_index, int)
            and isinstance(input_len, int)
        ):
            self.boundary_overrun_index = (
                ccheck_query_sink_result.boundary_overrun_index
            )
            self.end_index = ccheck_query_sink_result.end_index
            self.input_boundary_index = ccheck_query_sink_result.input_boundary_index
            self.start_index = ccheck_query_sink_result.start_index
            self.input_query = sql_query
            self.input_index = input_index
            self.input_len = input_len


def initialize_input_tracing():
    """
    This function is necessary because we only import agent_lib.modules.LIB_CONTRAST if
    we need to initialize agent-lib.
    """
    if not agent_lib.modules.is_initialized():
        return

    agent_lib.modules.LIB_CONTRAST.evaluate_header_input.argtypes = (
        ctypes.c_char_p,
        ctypes.c_char_p,
        ctypes.c_longlong,
        ctypes.c_longlong,
        ctypes.POINTER(ctypes.c_size_t),
        ctypes.POINTER(ctypes.POINTER(agent_lib.modules.CONSTANTS.CEvalResult)),
    )

    agent_lib.modules.LIB_CONTRAST.check_sql_injection_query.argtypes = (
        ctypes.c_uint32,
        ctypes.c_uint32,
        ctypes.c_uint32,
        ctypes.c_char_p,
        ctypes.POINTER(ctypes.POINTER(CCheckQuerySinkResult)),
    )

    agent_lib.modules.LIB_CONTRAST.evaluate_header_input.restype = ctypes.c_int
    agent_lib.modules.LIB_CONTRAST.check_sql_injection_query.restype = ctypes.c_int


def evaluate_header_input(header_name, header_value, rules, prefer_worth_watching):
    evaluations = []

    if not agent_lib.modules.is_initialized():
        return evaluations

    if rules == 0:
        return evaluations

    def is_valid_return(code):
        return code == 0

    name = ctypes.c_char_p(bytes(str(header_name), "utf8"))
    value = ctypes.c_char_p(bytes(str(header_value), "utf8"))
    results_len = ctypes.c_size_t()
    results = ctypes.POINTER(agent_lib.modules.CONSTANTS.CEvalResult)()

    ret = agent_lib.call(
        agent_lib.modules.LIB_CONTRAST.evaluate_header_input,
        is_valid_return,
        name,
        value,
        rules,
        prefer_worth_watching,
        ctypes.byref(results_len),
        ctypes.byref(results),
    )

    map_result_and_free_eval_result(
        ret,
        results,
        results_len,
        header_name,
        header_value,
        is_valid_return,
        evaluations,
        prefer_worth_watching,
    )
    return evaluations


def evaluate_input_by_type(input_type, input_value, rules, prefer_worth_watching):
    evaluations = []

    if not agent_lib.modules.is_initialized():
        return evaluations

    if rules == 0:
        return evaluations

    def is_valid_return(code):
        return code == 0

    # TODO: PYT-2826
    if not isinstance(input_value, str):
        name = ctypes.c_char_p(bytes(str(input_value), "utf8"))
    else:
        name = ctypes.c_char_p(bytes(input_value, "utf8"))
    value = ctypes.c_long(input_type)
    results_len = ctypes.c_size_t()
    results = ctypes.POINTER(agent_lib.modules.CONSTANTS.CEvalResult)()

    ret = agent_lib.call(
        agent_lib.modules.LIB_CONTRAST.evaluate_input,
        is_valid_return,
        name,
        value,
        rules,
        prefer_worth_watching,
        ctypes.byref(results_len),
        ctypes.byref(results),
    )

    map_result_and_free_eval_result(
        ret,
        results,
        results_len,
        name,
        input_value,
        is_valid_return,
        evaluations,
        prefer_worth_watching,
    )
    return evaluations


def check_method_tampering(input_value, prefer_worth_watching):
    evaluations = []

    if not agent_lib.modules.is_initialized():
        return evaluations

    def is_valid_return(code):
        return 1 >= code >= 0

    name = ctypes.c_char_p(bytes(input_value, "utf8"))

    ret = agent_lib.call(
        agent_lib.modules.LIB_CONTRAST.is_method_tampering,
        is_valid_return,
        name,
    )

    if ret != 1:
        return []

    c_eval_res = agent_lib.modules.CONSTANTS.CEvalResult()
    c_eval_res.rule_id = agent_lib.modules.CONSTANTS.RuleType.get("method-tampering")
    c_eval_res.input_type = agent_lib.modules.CONSTANTS.InputType.get("Method")
    c_eval_res.score = 100

    return [
        InputAnalysisResult(input_value, input_value, c_eval_res, prefer_worth_watching)
    ]


def check_sql_injection_query(
    user_input_start_index, user_input_len, db_type, built_sql_query
):
    if not agent_lib.modules.is_initialized():
        return None

    def is_valid_return(code):
        return -1 <= code <= 0

    input_index = ctypes.c_uint32(user_input_start_index)
    input_len = ctypes.c_uint32(user_input_len)
    db_type = ctypes.c_uint32(db_type)
    sql_query = ctypes.c_char_p(bytes(built_sql_query, "utf8"))
    results = ctypes.POINTER(CCheckQuerySinkResult)()

    ret = agent_lib.call(
        agent_lib.modules.LIB_CONTRAST.check_sql_injection_query,
        is_valid_return,
        input_index,
        input_len,
        db_type,
        sql_query,
        ctypes.byref(results),
    )

    evaluation = map_result_and_free_check_query_sink_result(
        ret,
        results,
        built_sql_query,
        user_input_start_index,
        user_input_len,
        is_valid_return,
    )
    return evaluation


def check_cmd_injection_query(user_input_start_index, user_input_len, user_input_txt):
    if not agent_lib.modules.is_initialized():
        return None

    def is_valid_return(code):
        return -1 <= code <= 0

    input_index = ctypes.c_uint32(user_input_start_index)
    input_len = ctypes.c_uint32(user_input_len)
    cmd_text = ctypes.c_char_p(bytes(user_input_txt, "utf8"))
    results = ctypes.POINTER(CCheckQuerySinkResult)()

    ret = agent_lib.call(
        agent_lib.modules.LIB_CONTRAST.check_cmd_injection_query,
        is_valid_return,
        input_index,
        input_len,
        cmd_text,
        ctypes.byref(results),
    )

    evaluation = map_result_and_free_check_query_sink_result(
        ret,
        results,
        user_input_txt,
        user_input_start_index,
        user_input_len,
        is_valid_return,
    )
    return evaluation


def map_result_and_free_eval_result(
    ret,
    results,
    results_len,
    name,
    value,
    is_valid_return,
    evaluations,
    prefer_worth_watching,
):
    if ret == 0 and bool(results) and results_len.value > 0:
        for i in range(results_len.value):
            evaluations.append(
                InputAnalysisResult(name, value, results[i], prefer_worth_watching)
            )

        # ctypes does not have OOR (original object return), it constructs a new,
        # equivalent object each time you retrieve an attribute.
        # So we can free right after we create our list
        agent_lib.call(
            agent_lib.modules.LIB_CONTRAST.free_eval_result,
            is_valid_return,
            results,
        )


def map_result_and_free_check_query_sink_result(
    ret, results, sql_query, input_index, input_len, is_valid_return
):
    if ret == 0 and bool(results):
        evaluation = SQLInjectionResult(
            sql_query, input_index, input_len, results.contents
        )

        # ctypes does not have OOR (original object return), it constructs a new,
        # equivalent object each time you retrieve an attribute.
        # So we can free right after we create our list
        agent_lib.call(
            agent_lib.modules.LIB_CONTRAST.free_check_query_sink_result,
            is_valid_return,
            results,
        )

        return evaluation
    return None



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Implements patches for the operator module

Our propagation rewrites are implemented in terms of these patches, so they
must always be enabled when Assess is enabled.
"""
import sys

from contrast_vendor.wrapt import register_post_import_hook

from contrast.agent import scope
from contrast.agent.assess.utils import is_trackable
from contrast.agent.assess.policy import string_propagation
from contrast.agent.policy import patch_manager
from contrast.patches.utils import analyze_policy, get_propagation_node
from contrast.utils.decorators import fail_quietly
from contrast.utils.patch_utils import build_and_apply_patch, wrap_and_watermark


@fail_quietly("failed to propagate through modulo in contrast__cformat__modulo")
def _propagate_cformat(result, format_str, args):
    propagation_func = None

    if isinstance(result, str):
        propagation_func = string_propagation.propagate_unicode_cformat
    elif isinstance(result, bytes):
        propagation_func = string_propagation.propagate_bytes_cformat
    elif isinstance(result, bytearray):
        # TODO: PYT-2709 - Update tests to verify bytearray is supported with and without funchook
        propagation_func = string_propagation.propagate_bytearray_cformat

    if propagation_func is None:
        return

    with scope.contrast_scope(), scope.propagation_scope():
        propagation_func(result, format_str, result, args, None)


def build_add_hook(original_func, patch_policy):
    policy_node = get_propagation_node(patch_policy)

    def add(wrapped, instance, args, kwargs):
        del instance

        with scope.contrast_scope():
            result = wrapped(*args, **kwargs)
        if not is_trackable(result) or scope.in_scope():
            return result

        analyze_policy(policy_node.name, result, args, kwargs)

        return result

    return wrap_and_watermark(original_func, add)


def build_mod_hook(original_func, patch_policy):
    # cformat is a bit of a special case so we don't use policy here
    del patch_policy

    def mod(wrapped, instance, args, kwargs):
        del instance

        with scope.contrast_scope():
            result = wrapped(*args, **kwargs)
        if not is_trackable(result) or scope.in_scope():
            return result

        _propagate_cformat(result, *args)

        return result

    return wrap_and_watermark(original_func, mod)


def patch_operator(operator_module):
    build_and_apply_patch(operator_module, "add", build_add_hook)
    build_and_apply_patch(operator_module, "iadd", build_add_hook)
    build_and_apply_patch(operator_module, "mod", build_mod_hook)


def register_patches():
    register_post_import_hook(patch_operator, "operator")


def reverse_patches():
    operator_module = sys.modules.get("operator")
    if not operator_module:
        return

    patch_manager.reverse_patches_by_owner(operator_module)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from collections import defaultdict
from importlib import import_module
import sys

from contrast.assess_extensions import smart_setattr
from contrast.utils import Namespace
from contrast.utils.object_utils import get_name

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class module(Namespace):
    # map from id(orig_attr_as_func) -> patch
    patch_map = {}
    # map from id(patch_as_func) -> orig_attr
    inverse_patch_map = {}
    # allows lookup of patches by owner ID
    # this is what enables reverse patching
    patches_by_owner = defaultdict(set)
    # allows lookup of patches by owner name
    patches_by_name = defaultdict(set)


def get_patch(obj_as_func):
    return module.patch_map.get(id(obj_as_func))


def patch(owner, name, patch=None):
    """
    Set attribute `name` of `owner` to `patch`.

    If `patch` is not provided, we look up the appropriate existing patch in
    the patch book and apply it. This behavior is used during repatching.

    :param owner: module or class that owns the original attribute
    :param name: str name of the attribute being patched
    :param patch: object replacing owner.name, or None to use an existing patch
    """
    orig_attr = getattr(owner, name, None)
    orig_attr_as_func = as_func(orig_attr)

    if orig_attr_as_func is None:
        # TODO: PYT-692 investigate unexpected patching
        logger.debug(
            "WARNING: failed to patch %s of %s: no such attribute", name, owner
        )
        return
    if patch is None:
        patch = get_patch(orig_attr_as_func)
        if patch is None:
            # TODO: PYT-692 investigate unexpected patching
            logger.debug(
                "WARNING: failed to repatch %s of %s: no entry in the patch map",
                name,
                owner,
            )
            return

    if id(orig_attr_as_func) in module.inverse_patch_map:
        # TODO: PYT-692 investigate unexpected patching
        logger.debug(
            "WARNING: patching over already patched method %s of %s", name, owner
        )

    smart_setattr(owner, name, patch)
    register_patch(owner, name, orig_attr)


def _reverse_patch(owner, name):
    """
    Restore a patched attribute back to its original

    :param owner: module or class that owns the attribute being reverse patched
    :param name: name of the attribute as a string
    """
    patch = getattr(owner, name)
    patch_as_func = as_func(patch)

    if not is_patched(patch):
        return

    orig_attr = module.inverse_patch_map[id(patch_as_func)]

    # TODO: PYT-2886 investigate why __new__ patches are not reversible.
    # If we encounter __new__ here, we are really just "pretending" to remove
    # the patch by deregistering it, but it does not actually get reversed.
    # This isn't ideal but the impact right now should be pretty minimal.
    if orig_attr is not object.__new__:
        smart_setattr(owner, name, orig_attr)

    _deregister_patch(patch_as_func, owner, name, orig_attr)


def reverse_patches_by_owner(owner):
    """
    Restore all patched attributes that belong to the owning module/class

    If the owner is a module, any patched classes in this module will not be
    automatically reversed by this method. For example, if the following are patched:

        foo.a
        foo.b
        foo.FooClass.foo_method

    in order to reverse the patches, it will be necessary to call this method twice:

        reverse_patches_by_owner(foo)
        reverse_patches_by_owner(foo.FooClass)

    :param owner: module or class that owns the attribute being reverse patched
    """
    if owner is None:
        return

    if id(owner) not in module.patches_by_owner:
        return

    for name in list(module.patches_by_owner[id(owner)]):
        _reverse_patch(owner, name)


def reverse_module_patches_by_name(module_name: str):
    """
    Reverse patches owned by module with given name

    If the module is not imported, this function has no effect

    :param module_name: name of the module
    """
    module = sys.modules.get(module_name)
    reverse_patches_by_owner(module)


def reverse_class_patches_by_name(module_name: str, class_name: str):
    module = sys.modules.get(module_name)
    reverse_patches_by_owner(getattr(module, class_name, None))


def reverse_all_patches():
    """
    Reverse every patch managed by the patch_manager.

    Currently, this only reverses direct references to attributes we've patched. It's
    still possible that this doesn't cover cases where we patched an extra reference to
    an attribute via repatching.
    """
    for owner_name in module.patches_by_name.copy():
        try:
            owner = import_module(owner_name)
        except ImportError:
            module_name, _, attr_name = owner_name.rpartition(".")
            owner = getattr(import_module(module_name), attr_name)

        reverse_patches_by_owner(owner)


def register_patch(owner, name, orig_attr):
    """
    Register patch in the patch map to prevent us from patching twice

    :param owner: module or class that owns the original function
    :param name: name of the patched attribute
    :param orig_attr: original attribute, which is being replaced
    """
    patch = getattr(owner, name)
    patch_as_func = as_func(patch)
    orig_as_func = as_func(orig_attr)

    if id(as_func(module.patch_map.get(id(orig_as_func)))) == id(patch_as_func):
        # this is the case for repatching: the original attribute already has a
        # registered patch and that patch matches the one we just applied to it
        return

    if patch_as_func is orig_as_func:
        # TODO: PYT-692 investigate unexpected patching
        logger.debug(
            "WARNING: attempt to register %s as a patch for itself - "
            "skipping patch map registration",
            orig_attr,
        )
        return

    module.patch_map[id(orig_as_func)] = patch
    module.inverse_patch_map[id(patch_as_func)] = orig_attr
    module.patches_by_owner[id(owner)].add(name)
    module.patches_by_name[get_name(owner)].add(name)


def _deregister_patch(patch_as_func, owner, name, orig_attr):
    """
    Remove the patch from all locations in the patch manager.
    """
    owner_name = get_name(owner)
    orig_as_func = as_func(orig_attr)
    module.patches_by_owner[id(owner)].discard(name)
    module.patches_by_name[owner_name].discard(name)
    # if by removing the `name` value from id(owner) set the set becomes
    # empty, remove the key from the dict, too.
    if not module.patches_by_owner[id(owner)]:
        del module.patches_by_owner[id(owner)]
        del module.patches_by_name[owner_name]

    # Safety check for the case where we actually have two different patches that
    # correspond to the same original function (e.g. some of the codecs patches). In
    # these cases, there are two entries in the inverse_patch_map, but only one in the
    # patch_map. This isn't ideal, but it prevents errors when reverse patching.
    if id(orig_as_func) in module.patch_map:
        del module.patch_map[id(orig_as_func)]

    del module.inverse_patch_map[id(patch_as_func)]

    from contrast.agent.policy.applicator import remove_patch_location

    remove_patch_location(owner, name)


def is_patched(attr):
    """
    If the given attribute is a key in the inverse patch map, it means that it is being
    used as a patch.

    :param attr: attribute in question
    :return: True if the attribute is a key in the inverse patch map, False otherwise
    """
    return id(as_func(attr)) in module.inverse_patch_map


def has_associated_patch(attr):
    """
    If we come across an attribute that's a value in the patch_map (a key in
    the inverse patch map), it should be patched. This is most useful during
    re-patching, where we might see an old reference to the unpatched original
    attribute.

    :param attr: attribute in question
    :return: True if the attribute is a key in the patch map, False otherwise
    """
    return id(as_func(attr)) in module.patch_map


def as_func(attr):
    """
    We can't trust the id of unbound methods. For example, if we have class Foo with
    instance method bar, Foo.bar returns a wrapper around the actual function object,
    and that wrapper may change between accesses to Foo.bar. This is accomplished with
    descriptors.

    Luckily, unbound methods should have a __func__ attribute, which references the
    raw underlying function. This value does not change, so we want to enter its id
    in the patch map.

    If we can't get a __func__ attribute, we try to use the _self_wrapper attribute,
    which our wrapt wrappers use
    """
    if not hasattr(attr, "__func__"):
        attr = getattr(attr, "_self_wrapper", attr)
    return getattr(attr, "__func__", attr)


def clear_patch_manager():
    """
    The nuclear option.

    The patch manager has trouble accounting for aliased patches. This leads to
    situations where patches have actually been reversed but there's still
    dangling metadata about aliased patches. Callers should perform due
    diligence to remove patches before calling this method.
    """
    module.patch_map.clear()
    module.inverse_patch_map.clear()
    module.patches_by_name.clear()
    module.patches_by_owner.clear()



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import print_function

import argparse
import os
import subprocess
import sys

INTERPRETER = os.path.realpath(sys.executable)


def fix_interpreter_permissions():
    args = _parse_args()
    _check_filetype()
    if not args.modify_interpreter:
        _warn_and_exit()
    _run()


def _parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--modify-interpreter",
        help="modify permissions for your python interpreter",
        action="store_true",
    )
    return parser.parse_args()


def _check_filetype():
    """
    Sanity check: if we don't see a Mach-O file format for the python
    interpreter, we may be about to make modifications with unknown effects.

    This check is not particularly robust, but it's about as good as we
    can do without significantly more effort (or a new dependency).
    """
    print(f"found interpreter: {INTERPRETER}")
    print("verifying interpreter filetype - expecting Mach-O")
    file_info = subprocess.check_output(["file", INTERPRETER]).decode().rstrip()
    print(file_info)

    if "Mach-O" not in file_info:
        raise RuntimeError("interpreter isn't a Mach-O executable")


def _warn_and_exit():
    warning_msg = """

################################### WARNING ###################################

This operation will modify the permissions of your python interpreter to allow
Contrast to apply its instrumentation. This is a workaround for an issue
specific to OSX python builds. `pyenv` is a popular tool that may lead to this
circumstance.

No known prebuilt installations of python have this issue. If you think you've
found one, please contact support@contrastsecurity.com.

This is a permanent in-place modification to your python interpreter. The real
path of your current interpreter is:

{}

If you wish to continue, please rerun this command with `--modify-interpreter`.
Exiting - no modifications performed.

###############################################################################

""".format(
        INTERPRETER
    )
    print(warning_msg)
    sys.exit(1)


def _run():
    print("found --modify-interpreter command line option")
    print(f"the following interpreter will be modified:\n{INTERPRETER}\n")

    byte_7 = subprocess.Popen(["echo", "-n", "\x07"], stdout=subprocess.PIPE)
    subprocess.check_call(
        [
            "dd",
            f"of={INTERPRETER}",
            "bs=1",
            "seek=160",
            "count=1",
            "conv=notrunc",
        ],
        stdin=byte_7.stdout,
    )

    print(
        "\nSuccess! Your interpreter is now compatible with Contrast's instrumentation."
    )



"""
Does parsing of ETag-related headers: If-None-Matches, If-Matches

Also If-Range parsing
"""

from contrast_vendor.webob.datetime_utils import (
    parse_date,
    serialize_date,
    )
from contrast_vendor.webob.descriptors import _rx_etag
from contrast_vendor.webob.util import header_docstring

__all__ = ['AnyETag', 'NoETag', 'ETagMatcher', 'IfRange', 'etag_property']


def etag_property(key, default, rfc_section, strong=True):
    doc = header_docstring(key, rfc_section)
    doc += "  Converts it as a Etag."

    def fget(req):
        value = req.environ.get(key)
        if not value:
            return default
        else:
            return ETagMatcher.parse(value, strong=strong)

    def fset(req, val):
        if val is None:
            req.environ[key] = None
        else:
            req.environ[key] = str(val)

    def fdel(req):
        del req.environ[key]

    return property(fget, fset, fdel, doc=doc)


class _AnyETag(object):
    """
    Represents an ETag of *, or a missing ETag when matching is 'safe'
    """

    def __repr__(self):
        return '<ETag *>'

    def __nonzero__(self):
        return False

    __bool__ = __nonzero__  # python 3

    def __contains__(self, other):
        return True

    def __str__(self):
        return '*'


AnyETag = _AnyETag()


class _NoETag(object):
    """
    Represents a missing ETag when matching is unsafe
    """

    def __repr__(self):
        return '<No ETag>'

    def __nonzero__(self):
        return False

    __bool__ = __nonzero__  # python 3

    def __contains__(self, other):
        return False

    def __str__(self):
        return ''


NoETag = _NoETag()


# TODO: convert into a simple tuple

class ETagMatcher(object):
    def __init__(self, etags):
        self.etags = etags

    def __contains__(self, other):
        return other in self.etags

    def __repr__(self):
        return '<ETag %s>' % (' or '.join(self.etags))

    @classmethod
    def parse(cls, value, strong=True):
        """
        Parse this from a header value
        """
        if value == '*':
            return AnyETag
        if not value:
            return cls([])
        matches = _rx_etag.findall(value)
        if not matches:
            return cls([value])
        elif strong:
            return cls([t for w, t in matches if not w])
        else:
            return cls([t for w, t in matches])

    def __str__(self):
        return ', '.join(map('"%s"'.__mod__, self.etags))


class IfRange(object):
    def __init__(self, etag):
        self.etag = etag

    @classmethod
    def parse(cls, value):
        """
        Parse this from a header value.
        """
        if not value:
            return cls(AnyETag)
        elif value.endswith(' GMT'):
            # Must be a date
            return IfRangeDate(parse_date(value))
        else:
            return cls(ETagMatcher.parse(value))

    def __contains__(self, resp):
        """
        Return True if the If-Range header matches the given etag or last_modified
        """
        return resp.etag_strong in self.etag

    def __nonzero__(self):
        return bool(self.etag)

    def __repr__(self):
        return '%s(%r)' % (
            self.__class__.__name__,
            self.etag
        )

    def __str__(self):
        return str(self.etag) if self.etag else ''

    __bool__ = __nonzero__  # python 3


class IfRangeDate(object):
    def __init__(self, date):
        self.date = date

    def __contains__(self, resp):
        last_modified = resp.last_modified
        return last_modified and (last_modified <= self.date)

    def __repr__(self):
        return '%s(%r)' % (
            self.__class__.__name__,
            self.date
        )

    def __str__(self):
        return serialize_date(self.date)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import functools
import inspect

try:
    from inspect import iscoroutinefunction
except ImportError:

    def iscoroutinefunction(f):
        return False


from contrast_vendor.wrapt import register_post_import_hook

import builtins
from contrast.agent.policy import patch_manager
from contrast.agent.policy import registry
from contrast.agent.protect.policy import apply_protect_patch
from contrast.agent.assess.policy.patches import (
    build_assess_method,
    build_assess_method_legacy,
    build_assess_classmethod,
    build_assess_deadzone,
    build_assess_property_fget,
    apply_cached_property,
    build_assess_async_method,
    build_assess_async_deadzone,
)
from contrast.agent.settings import Settings
from contrast.utils.patch_utils import repatch_module
from contrast.utils.safe_import import safe_import


from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


# Maintains a record of all patch locations that have been applied
PATCH_LOCATIONS = set()


@functools.lru_cache(maxsize=1)
def _get_cached_property():
    return safe_import("werkzeug.utils.cached_property")


def remove_patch_location(owner, name):
    """
    Remove patch located at owner (class or module) and name
    from patch locations.
    """
    if inspect.ismodule(owner):
        module_name = "BUILTIN" if owner is builtins else owner.__name__
        policy_patch_name = f"{module_name}.{name}"
    elif inspect.isclass(owner):
        policy_patch_name = f"{owner.__module__}.{owner.__name__}.{name}"

    if policy_patch_name in PATCH_LOCATIONS:
        PATCH_LOCATIONS.remove(policy_patch_name)


def apply_assess_patch(patch_site, patch_policy):
    """
    Store the original method implementation under our custom "cs_assess_{name}" so we
    can call the old method from the new method in order to return the same result

    :param cls_or_module: Class or module to patch
    :param patch_policy: PatchLocationPolicy containing all policy nodes for this patch
    :param methods: methods in cls_or_module
    :param node_method: tuple of string and actual method
        ('get_raw_uri', <unbound method django.http.request.HttpRequest.get_raw_uri>)

    Static methods and class methods are implemented as descriptors. When one of these
    methods is called on the parent object, it uses the descriptor protocol, which
    means that the underlying function objects are not seen directly by the caller.

    In order to access the staticmethod/classmethod objects directly, they must be
    accessed via the parent object's __dict__ attribute.

    For example,

    class D:
        @staticmethod
        def sm():
            pass

    If I do this:
    D.sm will call __get__ under the hood to get the function definition (this is
    what's wrapped inside the staticmethod or classmethod obj in this case)

    If we do this:
    D.__dict__['sm'] we get the underlying staticmethod/classmethod object. This is
    required because we need to do a type check in order to replace the original obj
    with our own function of the correct type.

    This shows examples of C code re-written in python for the __get__ desc
    implementation for classmethod, staticmethod and function
    https://docs.python.org/2.7/howto/descriptor.html#functions-and-methods

    For additional details on PY3:
    https://docs.python.org/3/howto/descriptor.html#functions-and-methods
    """
    logger.debug("Applying assess patch to %s", patch_policy.name)

    method_name = patch_policy.method_name
    has_key = hasattr(patch_site, "__dict__") and method_name in patch_site.__dict__

    static_method = has_key and isinstance(
        patch_site.__dict__[method_name], staticmethod
    )
    class_method = has_key and isinstance(patch_site.__dict__[method_name], classmethod)

    old_method = _get_original_method(
        patch_policy, patch_site, method_name, static_method, class_method
    )

    cached_property = _get_cached_property()
    if isinstance(old_method, property):
        # newer versions of werkzeug cached_property inherit from property
        # so these are handled here
        fget = build_assess_property_fget(old_method, patch_policy)
        new_method = property(fget=fget, fset=old_method.fset, fdel=old_method.fdel)
    elif cached_property is not None and isinstance(old_method, cached_property):
        # Handle the case in older versions of werkzeug where cached_propery does
        # not inherit from property.
        return apply_cached_property(patch_site, patch_policy, method_name, old_method)
    elif class_method:
        new_method = classmethod(build_assess_classmethod(old_method, patch_policy))
    elif static_method:
        new_method = staticmethod(build_assess_classmethod(old_method, patch_policy))
    elif iscoroutinefunction(old_method):
        new_method = (
            build_assess_async_deadzone(old_method, patch_policy)
            if patch_policy.is_deadzone
            else build_assess_async_method(old_method, patch_policy)
        )
    elif patch_policy.module in [
        "codecs",  # TODO: PYT-1975
        "falcon.util.uri",  # TODO: PYT-1976
    ]:
        new_method = build_assess_method_legacy(old_method, patch_policy)
    else:
        # We only support deadzones for methods/functions right now.
        if patch_policy.is_deadzone and not patch_policy.deadzone_enabled:
            return False

        new_method = (
            build_assess_deadzone(old_method, patch_policy)
            if patch_policy.is_deadzone
            else build_assess_method(old_method, patch_policy)
        )

    # OPTIMIZATION: If we know this patch location is only one type (trigger,
    # propagation, source) and only one node, then we can assign that analysis code here
    # at patch-time and save time when patch is called. Most patches will benefit
    # from optimization as very few functions are more than one thing (example:
    # re.split is both a trigger for redos rule and a propagator).
    if len(patch_policy.all_nodes) == 1:
        patch_policy.assign_analysis_func()

    try:
        patch_manager.patch(patch_site, method_name, new_method)
        logger.debug("added patch to %s.%s", patch_site, method_name)
    except Exception as e:
        logger.debug(
            "unable to patch %s method of %s: %s", method_name, patch_site, str(e)
        )
        return False

    return True


def _get_original_method(
    patch_policy, patch_site, method_name, static_method, class_method
):
    # Need to make sure we get original static/class function
    if static_method or class_method:
        old_method = patch_site.__dict__[method_name]
    else:
        # get old function/property/method
        try:
            old_method = getattr(patch_site, patch_policy.method_name)
        except AttributeError:
            # Some python/framework versions or platforms will not have certain
            # methods/funcs defined so we skip them here. This isn't a failure
            # so logging is not necessary.
            return False

    return old_method


def apply_module_patches(module):
    """
    Apply patches to all methods and functions in a module as dictated by policy
    """
    logger.debug("Running import hook for %s", module.__name__)

    module_name = "BUILTIN" if module is builtins else module.__name__

    module_policy = registry.get_policy_by_module(module_name)
    if module_policy is None:
        logger.debug("WARNING: No module policy found for %s", module_name)
        return

    settings = Settings()

    for patch_policy in module_policy:
        if patch_policy.name in PATCH_LOCATIONS:
            continue

        # If the module has no policy nodes, or if none of the nodes are policy
        # patches, then there's nothing to do here.
        if not patch_policy.has_patches:
            continue

        # If we want to patch with protect but this patch policy location does not
        # enable protect patching, move on.
        if settings.is_protect_enabled() and not patch_policy.is_protect_mode:
            continue

        if patch_policy.class_name:
            patch_site = getattr(module, patch_policy.class_name, None)
            if patch_site is None:
                continue
        else:
            patch_site = module

        patch_to_apply = (
            apply_protect_patch if settings.is_protect_enabled() else apply_assess_patch
        )
        try:
            patch_to_apply(patch_site, patch_policy)
        except Exception:
            logger.debug("Failed to apply patch for %s", patch_policy.name)

        PATCH_LOCATIONS.add(patch_policy.name)

    # It's possible that the current module contains multiple references to the
    # function we replaced, but that only one of them is represented in policy. We do a
    # quick repatching pass over the current module here to make sure we cover all of
    # the references that may have been missed.
    repatch_module(module)

    # EDGE CASE PYT-1065: Werkzeug==0.16.x
    # This version of Werkzeug keeps a reference to the module in _real_module
    # which requires us to repatch functions in this second reference, too.
    if hasattr(module, "_real_module"):
        repatch_module(module._real_module)


def register_import_hooks(protect_mode=False):
    """
    Use policy to register import hooks for each module that requires patches

    If protect_mode, patch only module patches with protect if trigger node has
    protect_mode: true.
    If not protect_mode, we will patch all module patches with assess.
    """
    modules_to_patch = registry.get_policy(protect=protect_mode)

    for module_name in modules_to_patch:
        # BUILTIN is a stand-in in policy for the builtins/__builtin__ module
        module_name = builtins.__name__ if module_name == "BUILTIN" else module_name
        logger.debug("Registering import hook for %s", module_name)
        register_post_import_hook(apply_module_patches, module_name)


def apply_patch_to_dynamic_property(class_to_patch, property_name, tags):
    """
    Take the property of a class we want to patch and:
        1. create a source node to store in policy
        2. patch the original property with our own code, including
            the policy instance with the new dynamic source.

    This means that the next time the cls.property is called,
    we will inject ourselves and run source policy.

    NOTE: adding the dynamic source to policy BEFORE patching is critical order,
    given that we have to patch with the policy instance that has this dynamic source.
    This could later be modified to not need this requirement.
    """
    module = class_to_patch.__module__
    class_name = class_to_patch.__name__

    patch_policy = registry.register_dynamic_source(
        module,
        class_name,
        property_name,
        tags,
        policy_patch=False,
    )

    orig_property = getattr(class_to_patch, property_name)

    def fset(cls_instance, value):
        # while good instinct would lead us to use setattr here instead of __dict__,
        # doing so does not work because we are in fact within a setter!
        cls_instance.__dict__[property_name] = value

    new_property = property(
        fget=build_assess_property_fget(orig_property, patch_policy),
        fset=fset,
        fdel=getattr(orig_property, "fdel", None),
    )

    patch_manager.patch(class_to_patch, property_name, new_property)

    return True



import binascii
import io
import os
import re
import sys
import tempfile
import mimetypes
try:
    import simplejson as json
except ImportError:
    import json
import warnings

from contrast_vendor.webob.acceptparse import (
    accept_charset_property,
    accept_encoding_property,
    accept_language_property,
    accept_property,
    )

from contrast_vendor.webob.cachecontrol import (
    CacheControl,
    serialize_cache_control,
    )

from contrast_vendor.webob.compat import (
    PY2,
    bytes_,
    native_,
    parse_qsl_text,
    reraise,
    text_type,
    url_encode,
    url_quote,
    url_unquote,
    quote_plus,
    urlparse,
    cgi_FieldStorage
    )

from contrast_vendor.webob.cookies import RequestCookies

from contrast_vendor.webob.descriptors import (
    CHARSET_RE,
    SCHEME_RE,
    converter,
    converter_date,
    environ_getter,
    environ_decoder,
    parse_auth,
    parse_int,
    parse_int_safe,
    parse_range,
    serialize_auth,
    serialize_if_range,
    serialize_int,
    serialize_range,
    upath_property,
    )

from contrast_vendor.webob.etag import (
    IfRange,
    AnyETag,
    NoETag,
    etag_property,
    )

from contrast_vendor.webob.headers import EnvironHeaders

from contrast_vendor.webob.multidict import (
    NestedMultiDict,
    MultiDict,
    NoVars,
    GetDict,
    )

__all__ = ['BaseRequest', 'Request', 'LegacyRequest']

class _NoDefault:
    def __repr__(self):
        return '(No Default)'
NoDefault = _NoDefault()

PATH_SAFE = "/~!$&'()*+,;=:@"

_LATIN_ENCODINGS = (
    'ascii', 'latin-1', 'latin', 'latin_1', 'l1', 'latin1',
    'iso-8859-1', 'iso8859_1', 'iso_8859_1', 'iso8859', '8859',
    )

class BaseRequest(object):
    # The limit after which request bodies should be stored on disk
    # if they are read in (under this, and the request body is stored
    # in memory):
    request_body_tempfile_limit = 10 * 1024

    _charset = None

    def __init__(self, environ, charset=None, unicode_errors=None,
                 decode_param_names=None, **kw):

        if type(environ) is not dict:
            raise TypeError(
                "WSGI environ must be a dict; you passed %r" % (environ,))

        if unicode_errors is not None:
            warnings.warn(
                "You unicode_errors=%r to the Request constructor.  Passing a "
                "``unicode_errors`` value to the Request is no longer "
                "supported in WebOb 1.2+.  This value has been ignored " % (
                    unicode_errors,),
                DeprecationWarning
                )

        if decode_param_names is not None:
            warnings.warn(
                "You passed decode_param_names=%r to the Request constructor. "
                "Passing a ``decode_param_names`` value to the Request "
                "is no longer supported in WebOb 1.2+.  This value has "
                "been ignored " % (decode_param_names,),
                DeprecationWarning
                )

        if not _is_utf8(charset):
            raise DeprecationWarning(
                "You passed charset=%r to the Request constructor. As of "
                "WebOb 1.2, if your application needs a non-UTF-8 request "
                "charset, please construct the request without a charset or "
                "with a charset of 'None',  then use ``req = "
                "req.decode(charset)``" % charset
            )

        d = self.__dict__
        d['environ'] = environ

        if kw:
            cls = self.__class__

            if 'method' in kw:
                # set method first, because .body setters
                # depend on it for checks
                self.method = kw.pop('method')

            for name, value in kw.items():
                if not hasattr(cls, name):
                    raise TypeError(
                        "Unexpected keyword: %s=%r" % (name, value))
                setattr(self, name, value)

    def encget(self, key, default=NoDefault, encattr=None):
        val = self.environ.get(key, default)
        if val is NoDefault:
            raise KeyError(key)
        if val is default:
            return default
        if not encattr:
            return val
        encoding = getattr(self, encattr)

        if PY2:
            return val.decode(encoding)

        if encoding in _LATIN_ENCODINGS: # shortcut
            return val
        return bytes_(val, 'latin-1').decode(encoding)

    def encset(self, key, val, encattr=None):
        if encattr:
            encoding = getattr(self, encattr)
        else:
            encoding = 'ascii'
        if PY2: # pragma: no cover
            self.environ[key] = bytes_(val, encoding)
        else:
            self.environ[key] = bytes_(val, encoding).decode('latin-1')

    @property
    def charset(self):
        if self._charset is None:
            charset = detect_charset(self._content_type_raw)
            if _is_utf8(charset):
                charset = 'UTF-8'
            self._charset = charset
        return self._charset

    @charset.setter
    def charset(self, charset):
        if _is_utf8(charset):
            charset = 'UTF-8'
        if charset != self.charset:
            raise DeprecationWarning("Use req = req.decode(%r)" % charset)

    def decode(self, charset=None, errors='strict'):
        charset = charset or self.charset
        if charset == 'UTF-8':
            return self
        # cookies and path are always utf-8
        t = Transcoder(charset, errors)

        new_content_type = CHARSET_RE.sub('; charset="UTF-8"',
                                          self._content_type_raw)
        content_type = self.content_type
        r = self.__class__(
            self.environ.copy(),
            query_string=t.transcode_query(self.query_string),
            content_type=new_content_type,
        )

        if content_type == 'application/x-www-form-urlencoded':
            r.body = bytes_(t.transcode_query(native_(self.body)))
            return r
        elif content_type != 'multipart/form-data':
            return r

        fs_environ = self.environ.copy()
        fs_environ.setdefault('CONTENT_LENGTH', '0')
        fs_environ['QUERY_STRING'] = ''
        if PY2:
            fs = cgi_FieldStorage(fp=self.body_file,
                                  environ=fs_environ,
                                  keep_blank_values=True)
        else:
            fs = cgi_FieldStorage(fp=self.body_file,
                                  environ=fs_environ,
                                  keep_blank_values=True,
                                  encoding=charset,
                                  errors=errors)

        fout = t.transcode_fs(fs, r._content_type_raw)

        # this order is important, because setting body_file
        # resets content_length
        r.body_file = fout
        r.content_length = fout.tell()
        fout.seek(0)
        return r

    # this is necessary for correct warnings depth for both
    # BaseRequest and Request (due to AdhocAttrMixin.__setattr__)
    _setattr_stacklevel = 2

    @property
    def body_file(self):
        """
            Input stream of the request (wsgi.input).
            Setting this property resets the content_length and seekable flag
            (unlike setting req.body_file_raw).
        """

        if not self.is_body_readable:
            return io.BytesIO()

        r = self.body_file_raw
        clen = self.content_length

        if not self.is_body_seekable and clen is not None:
            # we need to wrap input in LimitedLengthFile
            # but we have to cache the instance as well
            # otherwise this would stop working
            # (.remaining counter would reset between calls):
            #   req.body_file.read(100)
            #   req.body_file.read(100)
            env = self.environ
            wrapped, raw = env.get('webob._body_file', (0, 0))

            if raw is not r:
                wrapped = LimitedLengthFile(r, clen)
                wrapped = io.BufferedReader(wrapped)
                env['webob._body_file'] = wrapped, r
            r = wrapped

        return r

    @body_file.setter
    def body_file(self, value):
        if isinstance(value, bytes):
            raise ValueError('Excepted fileobj but received bytes.')

        self.content_length = None
        self.body_file_raw = value
        self.is_body_seekable = False
        self.is_body_readable = True

    @body_file.deleter
    def body_file(self):
        self.body = b''

    body_file_raw = environ_getter('wsgi.input')

    @property
    def body_file_seekable(self):
        """
            Get the body of the request (wsgi.input) as a seekable file-like
            object. Middleware and routing applications should use this
            attribute over .body_file.

            If you access this value, CONTENT_LENGTH will also be updated.
        """
        if not self.is_body_seekable:
            self.make_body_seekable()
        return self.body_file_raw

    url_encoding = environ_getter('webob.url_encoding', 'UTF-8')
    scheme = environ_getter('wsgi.url_scheme')
    method = environ_getter('REQUEST_METHOD', 'GET')
    http_version = environ_getter('SERVER_PROTOCOL')
    content_length = converter(
        environ_getter('CONTENT_LENGTH', None, '14.13'),
        parse_int_safe, serialize_int, 'int')
    remote_user = environ_getter('REMOTE_USER', None)
    remote_host = environ_getter('REMOTE_HOST', None)
    remote_addr = environ_getter('REMOTE_ADDR', None)
    query_string = environ_getter('QUERY_STRING', '')
    server_name = environ_getter('SERVER_NAME')
    server_port = converter(
        environ_getter('SERVER_PORT'),
        parse_int, serialize_int, 'int')

    script_name = environ_decoder('SCRIPT_NAME', '', encattr='url_encoding')
    path_info = environ_decoder('PATH_INFO', encattr='url_encoding')

    # bw compat
    uscript_name = script_name
    upath_info = path_info

    _content_type_raw = environ_getter('CONTENT_TYPE', '')

    def _content_type__get(self):
        """Return the content type, but leaving off any parameters (like
        charset, but also things like the type in ``application/atom+xml;
        type=entry``)

        If you set this property, you can include parameters, or if
        you don't include any parameters in the value then existing
        parameters will be preserved.
        """
        return self._content_type_raw.split(';', 1)[0]
    def _content_type__set(self, value=None):
        if value is not None:
            value = str(value)
            if ';' not in value:
                content_type = self._content_type_raw
                if ';' in content_type:
                    value += ';' + content_type.split(';', 1)[1]
        self._content_type_raw = value

    content_type = property(_content_type__get,
                            _content_type__set,
                            _content_type__set,
                            _content_type__get.__doc__)

    _headers = None

    def _headers__get(self):
        """
        All the request headers as a case-insensitive dictionary-like
        object.
        """
        if self._headers is None:
            self._headers = EnvironHeaders(self.environ)
        return self._headers

    def _headers__set(self, value):
        self.headers.clear()
        self.headers.update(value)

    headers = property(_headers__get, _headers__set, doc=_headers__get.__doc__)

    @property
    def client_addr(self):
        """
        The effective client IP address as a string.  If the
        ``HTTP_X_FORWARDED_FOR`` header exists in the WSGI environ, this
        attribute returns the client IP address present in that header
        (e.g. if the header value is ``192.168.1.1, 192.168.1.2``, the value
        will be ``192.168.1.1``). If no ``HTTP_X_FORWARDED_FOR`` header is
        present in the environ at all, this attribute will return the value
        of the ``REMOTE_ADDR`` header.  If the ``REMOTE_ADDR`` header is
        unset, this attribute will return the value ``None``.

        .. warning::

           It is possible for user agents to put someone else's IP or just
           any string in ``HTTP_X_FORWARDED_FOR`` as it is a normal HTTP
           header. Forward proxies can also provide incorrect values (private
           IP addresses etc).  You cannot "blindly" trust the result of this
           method to provide you with valid data unless you're certain that
           ``HTTP_X_FORWARDED_FOR`` has the correct values.  The WSGI server
           must be behind a trusted proxy for this to be true.
        """
        e = self.environ
        xff = e.get('HTTP_X_FORWARDED_FOR')
        if xff is not None:
            addr = xff.split(',')[0].strip()
        else:
            addr = e.get('REMOTE_ADDR')
        return addr

    @property
    def host_port(self):
        """
        The effective server port number as a string.  If the ``HTTP_HOST``
        header exists in the WSGI environ, this attribute returns the port
        number present in that header. If the ``HTTP_HOST`` header exists but
        contains no explicit port number: if the WSGI url scheme is "https" ,
        this attribute returns "443", if the WSGI url scheme is "http", this
        attribute returns "80" .  If no ``HTTP_HOST`` header is present in
        the environ at all, this attribute will return the value of the
        ``SERVER_PORT`` header (which is guaranteed to be present).
        """
        e = self.environ
        host = e.get('HTTP_HOST')
        if host is not None:
            if ':' in host and host[-1] != ']':
                host, port = host.rsplit(':', 1)
            else:
                url_scheme = e['wsgi.url_scheme']
                if url_scheme == 'https':
                    port = '443'
                else:
                    port = '80'
        else:
            port = e['SERVER_PORT']
        return port

    @property
    def host_url(self):
        """
        The URL through the host (no path)
        """
        e = self.environ
        scheme = e.get('wsgi.url_scheme')
        url = scheme + '://'
        host = e.get('HTTP_HOST')
        if host is not None:
            if ':' in host and host[-1] != ']':
                host, port = host.rsplit(':', 1)
            else:
                port = None
        else:
            host = e.get('SERVER_NAME')
            port = e.get('SERVER_PORT')
        if scheme == 'https':
            if port == '443':
                port = None
        elif scheme == 'http':
            if port == '80':
                port = None
        url += host
        if port:
            url += ':%s' % port
        return url

    @property
    def application_url(self):
        """
        The URL including SCRIPT_NAME (no PATH_INFO or query string)
        """
        bscript_name = bytes_(self.script_name, self.url_encoding)
        return self.host_url + url_quote(bscript_name, PATH_SAFE)

    @property
    def path_url(self):
        """
        The URL including SCRIPT_NAME and PATH_INFO, but not QUERY_STRING
        """
        bpath_info = bytes_(self.path_info, self.url_encoding)
        return self.application_url + url_quote(bpath_info, PATH_SAFE)

    @property
    def path(self):
        """
        The path of the request, without host or query string
        """
        bscript = bytes_(self.script_name, self.url_encoding)
        bpath = bytes_(self.path_info, self.url_encoding)
        return url_quote(bscript, PATH_SAFE) + url_quote(bpath, PATH_SAFE)

    @property
    def path_qs(self):
        """
        The path of the request, without host but with query string
        """
        path = self.path
        qs = self.environ.get('QUERY_STRING')
        if qs:
            path += '?' + qs
        return path

    @property
    def url(self):
        """
        The full request URL, including QUERY_STRING
        """
        url = self.path_url
        qs = self.environ.get('QUERY_STRING')
        if qs:
            url += '?' + qs
        return url

    def relative_url(self, other_url, to_application=False):
        """
        Resolve other_url relative to the request URL.

        If ``to_application`` is True, then resolve it relative to the
        URL with only SCRIPT_NAME
        """
        if to_application:
            url = self.application_url
            if not url.endswith('/'):
                url += '/'
        else:
            url = self.path_url
        return urlparse.urljoin(url, other_url)

    def path_info_pop(self, pattern=None):
        """
        'Pops' off the next segment of PATH_INFO, pushing it onto
        SCRIPT_NAME, and returning the popped segment.  Returns None if
        there is nothing left on PATH_INFO.

        Does not return ``''`` when there's an empty segment (like
        ``/path//path``); these segments are just ignored.

        Optional ``pattern`` argument is a regexp to match the return value
        before returning. If there is no match, no changes are made to the
        request and None is returned.
        """
        path = self.path_info
        if not path:
            return None
        slashes = ''
        while path.startswith('/'):
            slashes += '/'
            path = path[1:]
        idx = path.find('/')
        if idx == -1:
            idx = len(path)
        r = path[:idx]
        if pattern is None or re.match(pattern, r):
            self.script_name += slashes + r
            self.path_info = path[idx:]
            return r

    def path_info_peek(self):
        """
        Returns the next segment on PATH_INFO, or None if there is no
        next segment.  Doesn't modify the environment.
        """
        path = self.path_info
        if not path:
            return None
        path = path.lstrip('/')
        return path.split('/', 1)[0]

    def _urlvars__get(self):
        """
        Return any *named* variables matched in the URL.

        Takes values from ``environ['wsgiorg.routing_args']``.
        Systems like ``routes`` set this value.
        """
        if 'paste.urlvars' in self.environ:
            return self.environ['paste.urlvars']
        elif 'wsgiorg.routing_args' in self.environ:
            return self.environ['wsgiorg.routing_args'][1]
        else:
            result = {}
            self.environ['wsgiorg.routing_args'] = ((), result)
            return result

    def _urlvars__set(self, value):
        environ = self.environ
        if 'wsgiorg.routing_args' in environ:
            environ['wsgiorg.routing_args'] = (
                    environ['wsgiorg.routing_args'][0], value)
            if 'paste.urlvars' in environ:
                del environ['paste.urlvars']
        elif 'paste.urlvars' in environ:
            environ['paste.urlvars'] = value
        else:
            environ['wsgiorg.routing_args'] = ((), value)

    def _urlvars__del(self):
        if 'paste.urlvars' in self.environ:
            del self.environ['paste.urlvars']
        if 'wsgiorg.routing_args' in self.environ:
            if not self.environ['wsgiorg.routing_args'][0]:
                del self.environ['wsgiorg.routing_args']
            else:
                self.environ['wsgiorg.routing_args'] = (
                        self.environ['wsgiorg.routing_args'][0], {})

    urlvars = property(_urlvars__get,
                       _urlvars__set,
                       _urlvars__del,
                       doc=_urlvars__get.__doc__)

    def _urlargs__get(self):
        """
        Return any *positional* variables matched in the URL.

        Takes values from ``environ['wsgiorg.routing_args']``.
        Systems like ``routes`` set this value.
        """
        if 'wsgiorg.routing_args' in self.environ:
            return self.environ['wsgiorg.routing_args'][0]
        else:
            # Since you can't update this value in-place, we don't need
            # to set the key in the environment
            return ()

    def _urlargs__set(self, value):
        environ = self.environ
        if 'paste.urlvars' in environ:
            # Some overlap between this and wsgiorg.routing_args; we need
            # wsgiorg.routing_args to make this work
            routing_args = (value, environ.pop('paste.urlvars'))
        elif 'wsgiorg.routing_args' in environ:
            routing_args = (value, environ['wsgiorg.routing_args'][1])
        else:
            routing_args = (value, {})
        environ['wsgiorg.routing_args'] = routing_args

    def _urlargs__del(self):
        if 'wsgiorg.routing_args' in self.environ:
            if not self.environ['wsgiorg.routing_args'][1]:
                del self.environ['wsgiorg.routing_args']
            else:
                self.environ['wsgiorg.routing_args'] = (
                        (), self.environ['wsgiorg.routing_args'][1])

    urlargs = property(_urlargs__get,
                       _urlargs__set,
                       _urlargs__del,
                       _urlargs__get.__doc__)

    @property
    def is_xhr(self):
        """Is X-Requested-With header present and equal to ``XMLHttpRequest``?

        Note: this isn't set by every XMLHttpRequest request, it is
        only set if you are using a Javascript library that sets it
        (or you set the header yourself manually).  Currently
        Prototype and jQuery are known to set this header."""
        return self.environ.get('HTTP_X_REQUESTED_WITH', '') == 'XMLHttpRequest'

    def _host__get(self):
        """Host name provided in HTTP_HOST, with fall-back to SERVER_NAME"""
        if 'HTTP_HOST' in self.environ:
            return self.environ['HTTP_HOST']
        else:
            return '%(SERVER_NAME)s:%(SERVER_PORT)s' % self.environ
    def _host__set(self, value):
        self.environ['HTTP_HOST'] = value
    def _host__del(self):
        if 'HTTP_HOST' in self.environ:
            del self.environ['HTTP_HOST']
    host = property(_host__get, _host__set, _host__del, doc=_host__get.__doc__)

    @property
    def domain(self):
        """ Returns the domain portion of the host value.  Equivalent to:

        .. code-block:: python

           domain = request.host
           if ':' in domain and domain[-1] != ']': # Check for ] because of IPv6
               domain = domain.rsplit(':', 1)[0]

        This will be equivalent to the domain portion of the ``HTTP_HOST``
        value in the environment if it exists, or the ``SERVER_NAME`` value in
        the environment if it doesn't.  For example, if the environment
        contains an ``HTTP_HOST`` value of ``foo.example.com:8000``,
        ``request.domain`` will return ``foo.example.com``.

        Note that this value cannot be *set* on the request.  To set the host
        value use :meth:`webob.request.Request.host` instead.
        """
        domain = self.host
        if ':' in domain and domain[-1] != ']':
            domain = domain.rsplit(':', 1)[0]
        return domain

    @property
    def body(self):
        """
        Return the content of the request body.
        """
        if not self.is_body_readable:
            return b''

        self.make_body_seekable() # we need this to have content_length
        r = self.body_file.read(self.content_length)
        self.body_file_raw.seek(0)
        return r

    @body.setter
    def body(self, value):
        if value is None:
            value = b''
        if not isinstance(value, bytes):
            raise TypeError("You can only set Request.body to bytes (not %r)"
                            % type(value))
        self.content_length = len(value)
        self.body_file_raw = io.BytesIO(value)
        self.is_body_seekable = True

    @body.deleter
    def body(self):
        self.body = b''

    def _json_body__get(self):
        """Access the body of the request as JSON"""
        return json.loads(self.body.decode(self.charset))

    def _json_body__set(self, value):
        self.body = json.dumps(value, separators=(',', ':')).encode(self.charset)

    def _json_body__del(self):
        del self.body

    json = json_body = property(_json_body__get, _json_body__set, _json_body__del)

    def _text__get(self):
        """
        Get/set the text value of the body
        """
        if not self.charset:
            raise AttributeError(
                "You cannot access Request.text unless charset is set")
        body = self.body
        return body.decode(self.charset)

    def _text__set(self, value):
        if not self.charset:
            raise AttributeError(
                "You cannot access Response.text unless charset is set")
        if not isinstance(value, text_type):
            raise TypeError(
                "You can only set Request.text to a unicode string "
                "(not %s)" % type(value))
        self.body = value.encode(self.charset)

    def _text__del(self):
        del self.body

    text = property(_text__get, _text__set, _text__del, doc=_text__get.__doc__)

    @property
    def POST(self):
        """
        Return a MultiDict containing all the variables from a form
        request. Returns an empty dict-like object for non-form requests.

        Form requests are typically POST requests, however any other
        requests with an appropriate Content-Type are also supported.
        """
        env = self.environ
        if 'webob._parsed_post_vars' in env:
            vars, body_file = env['webob._parsed_post_vars']
            if body_file is self.body_file_raw:
                return vars
        content_type = self.content_type
        if ((self.method != 'POST' and not content_type)
            or content_type not in
                ('',
                 'application/x-www-form-urlencoded',
                 'multipart/form-data')
                 ):
            # Not an HTML form submission
            return NoVars('Not an HTML form submission (Content-Type: %s)'
                          % content_type)
        self._check_charset()

        self.make_body_seekable()
        self.body_file_raw.seek(0)

        fs_environ = env.copy()
        # FieldStorage assumes a missing CONTENT_LENGTH, but a
        # default of 0 is better:
        fs_environ.setdefault('CONTENT_LENGTH', '0')
        fs_environ['QUERY_STRING'] = ''
        if PY2:
            fs = cgi_FieldStorage(
                fp=self.body_file,
                environ=fs_environ,
                keep_blank_values=True)
        else:
            fs = cgi_FieldStorage(
                fp=self.body_file,
                environ=fs_environ,
                keep_blank_values=True,
                encoding='utf8')

        self.body_file_raw.seek(0)
        vars = MultiDict.from_fieldstorage(fs)
        env['webob._parsed_post_vars'] = (vars, self.body_file_raw)
        return vars

    @property
    def GET(self):
        """
        Return a MultiDict containing all the variables from the
        QUERY_STRING.
        """
        env = self.environ
        source = env.get('QUERY_STRING', '')
        if 'webob._parsed_query_vars' in env:
            vars, qs = env['webob._parsed_query_vars']
            if qs == source:
                return vars

        data = []
        if source:
            # this is disabled because we want to access req.GET
            # for text/plain; charset=ascii uploads for example
            #self._check_charset()
            data = parse_qsl_text(source)
            #d = lambda b: b.decode('utf8')
            #data = [(d(k), d(v)) for k,v in data]
        vars = GetDict(data, env)
        env['webob._parsed_query_vars'] = (vars, source)
        return vars

    def _check_charset(self):
        if self.charset != 'UTF-8':
            raise DeprecationWarning(
                "Requests are expected to be submitted in UTF-8, not %s. "
                "You can fix this by doing req = req.decode('%s')" % (
                    self.charset, self.charset)
            )

    @property
    def params(self):
        """
        A dictionary-like object containing both the parameters from
        the query string and request body.
        """
        params = NestedMultiDict(self.GET, self.POST)
        return params

    @property
    def cookies(self):
        """
        Return a dictionary of cookies as found in the request.
        """
        return RequestCookies(self.environ)

    @cookies.setter
    def cookies(self, val):
        self.environ.pop('HTTP_COOKIE', None)
        r = RequestCookies(self.environ)
        r.update(val)

    def copy(self):
        """
        Copy the request and environment object.

        This only does a shallow copy, except of wsgi.input
        """
        self.make_body_seekable()
        env = self.environ.copy()
        new_req = self.__class__(env)
        new_req.copy_body()
        return new_req

    def copy_get(self):
        """
        Copies the request and environment object, but turning this request
        into a GET along the way.  If this was a POST request (or any other
        verb) then it becomes GET, and the request body is thrown away.
        """
        env = self.environ.copy()
        return self.__class__(env, method='GET', content_type=None,
                              body=b'')

    # webob.is_body_seekable marks input streams that are seekable
    # this way we can have seekable input without testing the .seek() method
    is_body_seekable = environ_getter('webob.is_body_seekable', False)

    @property
    def is_body_readable(self):
        """
        webob.is_body_readable is a flag that tells us that we can read the
        input stream even though CONTENT_LENGTH is missing.
        """

        clen = self.content_length

        if clen is not None and clen != 0:
            return True
        elif clen is None:
            # Rely on the special flag that signifies that either Chunked
            # Encoding is allowed (and works) or we have replaced
            # self.body_file with something that is readable and EOF's
            # correctly.
            return self.environ.get(
                'wsgi.input_terminated',
                # For backwards compatibility, we fall back to checking if
                # webob.is_body_readable is set in the environ
                self.environ.get(
                    'webob.is_body_readable',
                    False
                )
            )

        return False

    @is_body_readable.setter
    def is_body_readable(self, flag):
        self.environ['wsgi.input_terminated'] = bool(flag)

    def make_body_seekable(self):
        """
        This forces ``environ['wsgi.input']`` to be seekable.
        That means that, the content is copied into a BytesIO or temporary
        file and flagged as seekable, so that it will not be unnecessarily
        copied again.

        After calling this method the .body_file is always seeked to the
        start of file and .content_length is not None.

        The choice to copy to BytesIO is made from
        ``self.request_body_tempfile_limit``
        """
        if self.is_body_seekable:
            self.body_file_raw.seek(0)
        else:
            self.copy_body()

    def copy_body(self):
        """
        Copies the body, in cases where it might be shared with another request
        object and that is not desired.

        This copies the body either into a BytesIO object (through setting
        req.body) or a temporary file.
        """

        if self.is_body_readable:
            # Before we copy, if we can, rewind the body file
            if self.is_body_seekable:
                self.body_file_raw.seek(0)

            tempfile_limit = self.request_body_tempfile_limit
            todo = self.content_length if self.content_length is not None else 65535

            newbody = b''
            fileobj = None
            input = self.body_file

            while todo > 0:
                data = input.read(min(todo, 65535))

                if not data and self.content_length is None:
                    # We attempted to read more data, but got none, break.
                    # This can happen if for instance we are reading as much as
                    # we can because we don't have a Content-Length...
                    break
                elif not data:
                    # We have a Content-Length and we attempted to read, but
                    # there was nothing more to read. Oh the humanity! This
                    # should rarely if never happen because self.body_file
                    # should be a LimitedLengthFile which should already have
                    # raised if there was less data than expected.
                    raise DisconnectionError(
                        "Client disconnected (%s more bytes were expected)" % todo
                    )

                if fileobj:
                    fileobj.write(data)
                else:
                    newbody += data

                    # When we have enough data that we need a tempfile, let's
                    # create one, then clear the temporary variable we were
                    # using
                    if len(newbody) > tempfile_limit:
                        fileobj = self.make_tempfile()
                        fileobj.write(newbody)
                        newbody = b''

                # Only decrement todo if Content-Length is set
                if self.content_length is not None:
                    todo -= len(data)

            if fileobj:
                # We apparently had enough data to need a file

                # Set the Content-Length to the amount of data that was just
                # written.
                self.content_length = fileobj.tell()

                # Seek it back to the beginning
                fileobj.seek(0)

                self.body_file_raw = fileobj

                # Allow it to be seeked in the future, so we don't need to copy
                # for things like .body
                self.is_body_seekable = True

                # Not strictly required since Content-Length is set
                self.is_body_readable = True
            else:
                # No file created, set the body and let it deal with creating
                # Content-Length and other vars.
                self.body = newbody
        else:
            # Always leave the request with a valid body, and this is pretty
            # cheap.
            self.body = b''

    def make_tempfile(self):
        """
            Create a tempfile to store big request body.
            This API is not stable yet. A 'size' argument might be added.
        """
        return tempfile.TemporaryFile()

    def remove_conditional_headers(self,
                                   remove_encoding=True,
                                   remove_range=True,
                                   remove_match=True,
                                   remove_modified=True):
        """
        Remove headers that make the request conditional.

        These headers can cause the response to be 304 Not Modified,
        which in some cases you may not want to be possible.

        This does not remove headers like If-Match, which are used for
        conflict detection.
        """
        check_keys = []
        if remove_range:
            check_keys += ['HTTP_IF_RANGE', 'HTTP_RANGE']
        if remove_match:
            check_keys.append('HTTP_IF_NONE_MATCH')
        if remove_modified:
            check_keys.append('HTTP_IF_MODIFIED_SINCE')
        if remove_encoding:
            check_keys.append('HTTP_ACCEPT_ENCODING')

        for key in check_keys:
            if key in self.environ:
                del self.environ[key]

    accept = accept_property()
    accept_charset = accept_charset_property()
    accept_encoding = accept_encoding_property()
    accept_language = accept_language_property()

    authorization = converter(
        environ_getter('HTTP_AUTHORIZATION', None, '14.8'),
        parse_auth, serialize_auth,
    )

    def _cache_control__get(self):
        """
        Get/set/modify the Cache-Control header (`HTTP spec section 14.9
        <http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.9>`_)
        """
        env = self.environ
        value = env.get('HTTP_CACHE_CONTROL', '')
        cache_header, cache_obj = env.get('webob._cache_control', (None, None))
        if cache_obj is not None and cache_header == value:
            return cache_obj
        cache_obj = CacheControl.parse(value,
                                       updates_to=self._update_cache_control,
                                       type='request')
        env['webob._cache_control'] = (value, cache_obj)
        return cache_obj

    def _cache_control__set(self, value):
        env = self.environ
        value = value or ''
        if isinstance(value, dict):
            value = CacheControl(value, type='request')
        if isinstance(value, CacheControl):
            str_value = str(value)
            env['HTTP_CACHE_CONTROL'] = str_value
            env['webob._cache_control'] = (str_value, value)
        else:
            env['HTTP_CACHE_CONTROL'] = str(value)
            env['webob._cache_control'] = (None, None)

    def _cache_control__del(self):
        env = self.environ
        if 'HTTP_CACHE_CONTROL' in env:
            del env['HTTP_CACHE_CONTROL']
        if 'webob._cache_control' in env:
            del env['webob._cache_control']

    def _update_cache_control(self, prop_dict):
        self.environ['HTTP_CACHE_CONTROL'] = serialize_cache_control(prop_dict)

    cache_control = property(_cache_control__get,
                             _cache_control__set,
                             _cache_control__del,
                             doc=_cache_control__get.__doc__)


    if_match = etag_property('HTTP_IF_MATCH', AnyETag, '14.24')
    if_none_match = etag_property('HTTP_IF_NONE_MATCH', NoETag, '14.26',
                                  strong=False)

    date = converter_date(environ_getter('HTTP_DATE', None, '14.8'))
    if_modified_since = converter_date(
                    environ_getter('HTTP_IF_MODIFIED_SINCE', None, '14.25'))
    if_unmodified_since = converter_date(
                    environ_getter('HTTP_IF_UNMODIFIED_SINCE', None, '14.28'))
    if_range = converter(
        environ_getter('HTTP_IF_RANGE', None, '14.27'),
        IfRange.parse, serialize_if_range, 'IfRange object')


    max_forwards = converter(
        environ_getter('HTTP_MAX_FORWARDS', None, '14.31'),
        parse_int, serialize_int, 'int')

    pragma = environ_getter('HTTP_PRAGMA', None, '14.32')

    range = converter(
        environ_getter('HTTP_RANGE', None, '14.35'),
        parse_range, serialize_range, 'Range object')

    referer = environ_getter('HTTP_REFERER', None, '14.36')
    referrer = referer

    user_agent = environ_getter('HTTP_USER_AGENT', None, '14.43')

    def __repr__(self):
        try:
            name = '%s %s' % (self.method, self.url)
        except KeyError:
            name = '(invalid WSGI environ)'
        msg = '<%s at 0x%x %s>' % (
            self.__class__.__name__,
            abs(id(self)), name)
        return msg

    def as_bytes(self, skip_body=False):
        """
            Return HTTP bytes representing this request.
            If skip_body is True, exclude the body.
            If skip_body is an integer larger than one, skip body
            only if its length is bigger than that number.
        """
        url = self.url
        host = self.host_url
        assert url.startswith(host)
        url = url[len(host):]
        parts = [bytes_('%s %s %s' % (self.method, url, self.http_version))]

        # acquire body before we handle headers so that
        # content-length will be set
        body = None
        if self.is_body_readable:
            if skip_body > 1:
                if len(self.body) > skip_body:
                    body = bytes_('<body skipped (len=%s)>' % len(self.body))
                else:
                    skip_body = False
            if not skip_body:
                body = self.body

        for k, v in sorted(self.headers.items()):
            header = bytes_('%s: %s'  % (k, v))
            parts.append(header)

        if body:
            parts.extend([b'', body])
        # HTTP clearly specifies CRLF
        return b'\r\n'.join(parts)

    def as_text(self):
        bytes = self.as_bytes()
        return bytes.decode(self.charset)

    __str__ = as_text

    @classmethod
    def from_bytes(cls, b):
        """
            Create a request from HTTP bytes data. If the bytes contain
            extra data after the request, raise a ValueError.
        """
        f = io.BytesIO(b)
        r = cls.from_file(f)
        if f.tell() != len(b):
            raise ValueError("The string contains more data than expected")
        return r

    @classmethod
    def from_text(cls, s):
        b = bytes_(s, 'utf-8')
        return cls.from_bytes(b)

    @classmethod
    def from_file(cls, fp):
        """Read a request from a file-like object (it must implement
        ``.read(size)`` and ``.readline()``).

        It will read up to the end of the request, not the end of the
        file (unless the request is a POST or PUT and has no
        Content-Length, in that case, the entire file is read).

        This reads the request as represented by ``str(req)``; it may
        not read every valid HTTP request properly.
        """
        start_line = fp.readline()
        is_text = isinstance(start_line, text_type)
        if is_text:
            crlf = '\r\n'
            colon = ':'
        else:
            crlf = b'\r\n'
            colon = b':'
        try:
            header = start_line.rstrip(crlf)
            method, resource, http_version = header.split(None, 2)
            method = native_(method, 'utf-8')
            resource = native_(resource, 'utf-8')
            http_version = native_(http_version, 'utf-8')
        except ValueError:
            raise ValueError('Bad HTTP request line: %r' % start_line)
        r = cls(environ_from_url(resource),
                http_version=http_version,
                method=method.upper()
                )
        del r.environ['HTTP_HOST']
        while 1:
            line = fp.readline()
            if not line.strip():
                # end of headers
                break
            hname, hval = line.split(colon, 1)
            hname = native_(hname, 'utf-8')
            hval = native_(hval, 'utf-8').strip()
            if hname in r.headers:
                hval = r.headers[hname] + ', ' + hval
            r.headers[hname] = hval

        clen = r.content_length
        if clen is None:
            body = fp.read()
        else:
            body = fp.read(clen)
        if is_text:
            body = bytes_(body, 'utf-8')
        r.body = body

        return r

    def call_application(self, application, catch_exc_info=False):
        """
        Call the given WSGI application, returning ``(status_string,
        headerlist, app_iter)``

        Be sure to call ``app_iter.close()`` if it's there.

        If catch_exc_info is true, then returns ``(status_string,
        headerlist, app_iter, exc_info)``, where the fourth item may
        be None, but won't be if there was an exception.  If you don't
        do this and there was an exception, the exception will be
        raised directly.
        """
        if self.is_body_seekable:
            self.body_file_raw.seek(0)
        captured = []
        output = []
        def start_response(status, headers, exc_info=None):
            if exc_info is not None and not catch_exc_info:
                reraise(exc_info)
            captured[:] = [status, headers, exc_info]
            return output.append
        app_iter = application(self.environ, start_response)
        if output or not captured:
            try:
                output.extend(app_iter)
            finally:
                if hasattr(app_iter, 'close'):
                    app_iter.close()
            app_iter = output
        if catch_exc_info:
            return (captured[0], captured[1], app_iter, captured[2])
        else:
            return (captured[0], captured[1], app_iter)

    # Will be filled in later:
    ResponseClass = None

    def send(self, application=None, catch_exc_info=False):
        """
        Like ``.call_application(application)``, except returns a
        response object with ``.status``, ``.headers``, and ``.body``
        attributes.

        This will use ``self.ResponseClass`` to figure out the class
        of the response object to return.

        If ``application`` is not given, this will send the request to
        ``self.make_default_send_app()``
        """
        if application is None:
            application = self.make_default_send_app()
        if catch_exc_info:
            status, headers, app_iter, exc_info = self.call_application(
                application, catch_exc_info=True)
            del exc_info
        else:
            status, headers, app_iter = self.call_application(
                application, catch_exc_info=False)
        return self.ResponseClass(
            status=status, headerlist=list(headers), app_iter=app_iter)

    get_response = send

    def make_default_send_app(self):
        global _client
        try:
            client = _client
        except NameError:
            from contrast_vendor.webob import client
            _client = client
        return client.send_request_app

    @classmethod
    def blank(cls, path, environ=None, base_url=None,
              headers=None, POST=None, **kw):
        """
        Create a blank request environ (and Request wrapper) with the
        given path (path should be urlencoded), and any keys from
        environ.

        The path will become path_info, with any query string split
        off and used.

        All necessary keys will be added to the environ, but the
        values you pass in will take precedence.  If you pass in
        base_url then wsgi.url_scheme, HTTP_HOST, and SCRIPT_NAME will
        be filled in from that value.

        Any extra keyword will be passed to ``__init__``.
        """
        env = environ_from_url(path)
        if base_url:
            scheme, netloc, path, query, fragment = urlparse.urlsplit(base_url)
            if query or fragment:
                raise ValueError(
                    "base_url (%r) cannot have a query or fragment"
                    % base_url)
            if scheme:
                env['wsgi.url_scheme'] = scheme
            if netloc:
                if ':' not in netloc:
                    if scheme == 'http':
                        netloc += ':80'
                    elif scheme == 'https':
                        netloc += ':443'
                    else:
                        raise ValueError(
                            "Unknown scheme: %r" % scheme)
                host, port = netloc.split(':', 1)
                env['SERVER_PORT'] = port
                env['SERVER_NAME'] = host
                env['HTTP_HOST'] = netloc
            if path:
                env['SCRIPT_NAME'] = url_unquote(path)
        if environ:
            env.update(environ)
        content_type = kw.get('content_type', env.get('CONTENT_TYPE'))
        if headers and 'Content-Type' in headers:
            content_type = headers['Content-Type']
        if content_type is not None:
            kw['content_type'] = content_type
        environ_add_POST(env, POST, content_type=content_type)
        obj = cls(env, **kw)
        if headers is not None:
            obj.headers.update(headers)
        return obj

class LegacyRequest(BaseRequest):
    uscript_name = upath_property('SCRIPT_NAME')
    upath_info = upath_property('PATH_INFO')

    def encget(self, key, default=NoDefault, encattr=None):
        val = self.environ.get(key, default)
        if val is NoDefault:
            raise KeyError(key)
        if val is default:
            return default
        return val

class AdhocAttrMixin(object):
    _setattr_stacklevel = 3

    def __setattr__(self, attr, value, DEFAULT=object()):
        if (getattr(self.__class__, attr, DEFAULT) is not DEFAULT or
                    attr.startswith('_')):
            object.__setattr__(self, attr, value)
        else:
            self.environ.setdefault('webob.adhoc_attrs', {})[attr] = value

    def __getattr__(self, attr, DEFAULT=object()):
        try:
            return self.environ['webob.adhoc_attrs'][attr]
        except KeyError:
            raise AttributeError(attr)

    def __delattr__(self, attr, DEFAULT=object()):
        if getattr(self.__class__, attr, DEFAULT) is not DEFAULT:
            return object.__delattr__(self, attr)
        try:
            del self.environ['webob.adhoc_attrs'][attr]
        except KeyError:
            raise AttributeError(attr)

class Request(AdhocAttrMixin, BaseRequest):
    """ The default request implementation """

def environ_from_url(path):
    if SCHEME_RE.search(path):
        scheme, netloc, path, qs, fragment = urlparse.urlsplit(path)
        if fragment:
            raise TypeError("Path cannot contain a fragment (%r)" % fragment)
        if qs:
            path += '?' + qs
        if ':' not in netloc:
            if scheme == 'http':
                netloc += ':80'
            elif scheme == 'https':
                netloc += ':443'
            else:
                raise TypeError("Unknown scheme: %r" % scheme)
    else:
        scheme = 'http'
        netloc = 'localhost:80'
    if path and '?' in path:
        path_info, query_string = path.split('?', 1)
        path_info = url_unquote(path_info)
    else:
        path_info = url_unquote(path)
        query_string = ''
    env = {
        'REQUEST_METHOD': 'GET',
        'SCRIPT_NAME': '',
        'PATH_INFO': path_info or '',
        'QUERY_STRING': query_string,
        'SERVER_NAME': netloc.split(':')[0],
        'SERVER_PORT': netloc.split(':')[1],
        'HTTP_HOST': netloc,
        'SERVER_PROTOCOL': 'HTTP/1.0',
        'wsgi.version': (1, 0),
        'wsgi.url_scheme': scheme,
        'wsgi.input': io.BytesIO(),
        'wsgi.errors': sys.stderr,
        'wsgi.multithread': False,
        'wsgi.multiprocess': False,
        'wsgi.run_once': False,
        #'webob.is_body_seekable': True,
    }
    return env


def environ_add_POST(env, data, content_type=None):
    if data is None:
        return
    elif isinstance(data, text_type):
        data = data.encode('ascii')
    if env['REQUEST_METHOD'] not in ('POST', 'PUT'):
        env['REQUEST_METHOD'] = 'POST'
    has_files = False
    if hasattr(data, 'items'):
        data = list(data.items())
        for k, v in data:
            if isinstance(v, (tuple, list)):
                has_files = True
                break
    if content_type is None:
        if has_files:
            content_type = 'multipart/form-data'
        else:
            content_type = 'application/x-www-form-urlencoded'
    if content_type.startswith('multipart/form-data'):
        if not isinstance(data, bytes):
            content_type, data = _encode_multipart(data, content_type)
    elif content_type.startswith('application/x-www-form-urlencoded'):
        if has_files:
            raise ValueError('Submiting files is not allowed for'
                             ' content type `%s`' % content_type)
        if not isinstance(data, bytes):
            data = url_encode(data)
    else:
        if not isinstance(data, bytes):
            raise ValueError('Please provide `POST` data as bytes'
                             ' for content type `%s`' % content_type)
    data = bytes_(data, 'utf8')
    env['wsgi.input'] = io.BytesIO(data)
    env['webob.is_body_seekable'] = True
    env['CONTENT_LENGTH'] = str(len(data))
    env['CONTENT_TYPE'] = content_type


#
# Helper classes and monkeypatching
#

class DisconnectionError(IOError):
    pass


class LimitedLengthFile(io.RawIOBase):
    def __init__(self, file, maxlen):
        self.file = file
        self.maxlen = maxlen
        self.remaining = maxlen

    def __repr__(self):
        return '<%s(%r, maxlen=%s)>' % (
            self.__class__.__name__,
            self.file,
            self.maxlen
        )

    def fileno(self):
        return self.file.fileno()

    @staticmethod
    def readable():
        return True

    def readinto(self, buff):
        if not self.remaining:
            return 0
        sz0 = min(len(buff), self.remaining)
        data = self.file.read(sz0)
        sz = len(data)
        self.remaining -= sz
        if sz < sz0 and self.remaining:
            raise DisconnectionError(
                "The client disconnected while sending the body "
                "(%d more bytes were expected)" % (self.remaining,)
            )
        buff[:sz] = data
        return sz


def _cgi_FieldStorage__repr__patch(self):
    """ monkey patch for FieldStorage.__repr__

    Unbelievably, the default __repr__ on FieldStorage reads
    the entire file content instead of being sane about it.
    This is a simple replacement that doesn't do that
    """
    if self.file:
        return "FieldStorage(%r, %r)" % (self.name, self.filename)
    return "FieldStorage(%r, %r, %r)" % (self.name, self.filename, self.value)

cgi_FieldStorage.__repr__ = _cgi_FieldStorage__repr__patch


class FakeCGIBody(io.RawIOBase):
    def __init__(self, vars, content_type):
        warnings.warn(
            "FakeCGIBody is no longer used by WebOb and will be removed from a future "
            "version of WebOb. If you require FakeCGIBody please make a copy into "
            "you own project",
            DeprecationWarning
            )

        if content_type.startswith('multipart/form-data'):
            if not _get_multipart_boundary(content_type):
                raise ValueError('Content-type: %r does not contain boundary'
                                 % content_type)
        self.vars = vars
        self.content_type = content_type
        self.file = None

    def __repr__(self):
        inner = repr(self.vars)
        if len(inner) > 20:
            inner = inner[:15] + '...' + inner[-5:]
        return '<%s at 0x%x viewing %s>' % (
            self.__class__.__name__,
            abs(id(self)), inner)

    def fileno(self):
        return None

    @staticmethod
    def readable():
        return True

    def readinto(self, buff):
        if self.file is None:
            if self.content_type.startswith('application/x-www-form-urlencoded'):
                data = '&'.join(
                    '%s=%s' % (
                        quote_plus(bytes_(k, 'utf8')),
                        quote_plus(bytes_(v, 'utf8'))
                    )
                    for k, v in self.vars.items()
                )
                self.file = io.BytesIO(bytes_(data))
            elif self.content_type.startswith('multipart/form-data'):
                self.file = _encode_multipart(
                    self.vars.items(),
                    self.content_type,
                    fout=io.BytesIO()
                )[1]
                self.file.seek(0)
            else:
                assert 0, ('Bad content type: %r' % self.content_type)
        return self.file.readinto(buff)

def _get_multipart_boundary(ctype):
    m = re.search(r'boundary=([^ ]+)', ctype, re.I)
    if m:
        return native_(m.group(1).strip('"'))

def _encode_multipart(vars, content_type, fout=None):
    """Encode a multipart request body into a string"""
    f = fout or io.BytesIO()
    w = f.write
    def wt(t):
        w(t.encode('utf8'))

    CRLF = b'\r\n'
    boundary = _get_multipart_boundary(content_type)
    if not boundary:
        boundary = native_(binascii.hexlify(os.urandom(10)))
        content_type += ('; boundary=%s' % boundary)
    for name, value in vars:
        w(b'--')
        wt(boundary)
        w(CRLF)
        wt('Content-Disposition: form-data')
        if name is not None:
            wt('; name="%s"' % name)
        filename = None
        if getattr(value, 'filename', None):
            filename = value.filename
        elif isinstance(value, (list, tuple)):
            filename, value = value
            if hasattr(value, 'read'):
                value = value.read()

        if filename is not None:
            wt('; filename="%s"' % filename)
            mime_type = mimetypes.guess_type(filename)[0]
        else:
            mime_type = None

        w(CRLF)

        # TODO: should handle value.disposition_options
        if getattr(value, 'type', None):
            wt('Content-type: %s' % value.type)
            if value.type_options:
                for ct_name, ct_value in sorted(value.type_options.items()):
                    wt('; %s="%s"' % (ct_name, ct_value))
            w(CRLF)
        elif mime_type:
            wt('Content-type: %s' % mime_type)
            w(CRLF)
        w(CRLF)
        if hasattr(value, 'value'):
            value = value.value
        if isinstance(value, bytes):
            w(value)
        else:
            wt(value)
        w(CRLF)
    wt('--%s--' % boundary)
    if fout:
        return content_type, fout
    else:
        return content_type, f.getvalue()

def detect_charset(ctype):
    m = CHARSET_RE.search(ctype)
    if m:
        return m.group(1).strip('"').strip()

def _is_utf8(charset):
    if not charset:
        return True
    else:
        return charset.lower().replace('-', '') == 'utf8'


class Transcoder(object):
    def __init__(self, charset, errors='strict'):
        self.charset = charset # source charset
        self.errors = errors # unicode errors
        self._trans = lambda b: b.decode(charset, errors).encode('utf8')

    def transcode_query(self, q):
        q_orig = q
        if '=' not in q:
            # this doesn't look like a form submission
            return q_orig

        if PY2:
            q = urlparse.parse_qsl(q, self.charset)
            t = self._trans
            q = [(t(k), t(v)) for k, v in q]
        else:
            q = list(parse_qsl_text(q, self.charset))

        return url_encode(q)

    def transcode_fs(self, fs, content_type):
        # transcode FieldStorage
        if PY2:
            def decode(b):
                if b is not None:
                    return b.decode(self.charset, self.errors)
                else:
                    return b
        else:
            def decode(b):
                return b

        data = []
        for field in fs.list or ():
            field.name = decode(field.name)
            if field.filename:
                field.filename = decode(field.filename)
                data.append((field.name, field))
            else:
                data.append((field.name, decode(field.value)))

        # TODO: transcode big requests to temp file
        content_type, fout = _encode_multipart(
            data,
            content_type,
            fout=io.BytesIO()
        )
        return fout



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.policy.registry import register_propagation_nodes


path_propagators = [
    {
        # os.path is just an alias for posixpath
        "module": "posixpath",
        "method_name": "basename",
        "source": "ARG_0,KWARG:p",
        "target": "RETURN",
        # We rely on rewrites to perform the actual propagation
        "action": "TAGGER",
        "tags": ["SAFE_PATH"],
    },
    {
        "module": "posix",
        "method_name": ["_path_normpath", "readlink"],
        "source": "ARG_0,KWARG:path",
        "target": "RETURN",
        "action": "SPLAT",
    },
    {
        "module": "urllib.parse",
        "method_name": ["quote", "quote_plus"],
        "source": "ARG_0,KWARG:string",
        "target": "RETURN",
        "action": "SPLAT",
        "tags": ["URL_ENCODED"],
    },
    {
        "module": "urllib.parse",
        "method_name": ["unquote", "unquote_plus"],
        "source": "ARG_0,KWARG:string",
        "target": "RETURN",
        "action": "SPLAT",
        "untags": ["URL_ENCODED"],
    },
    {
        "module": "urllib3.util.url",
        "method_name": ["_encode_invalid_chars"],
        "source": "ARG_0,KWARG:component",
        "target": "RETURN",
        "action": "SPLAT",
    },
]


register_propagation_nodes(path_propagators)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from copy import copy
import time

import contrast
from contrast.agent.assess.contrast_event import ContrastEvent
from contrast.agent.assess.tag import (
    ABOVE,
    BELOW,
    HIGH_SPAN,
    LOW_SPAN,
    Tag,
    WITHIN,
    WITHOUT,
)
from contrast.api.trace_event import TraceTaintRange
from contrast.utils.assess.tag_utils import merge_tags, ordered_merge


class Properties(object):
    def __init__(self, origin):
        self.origin = origin
        self._timestamp = time.time()
        self.reset()

    @property
    def timestamp(self):
        # NOTE: timestamp must not be modified. This affects string tracker age-off.
        return self._timestamp

    def reset(self):
        self.tags = {}
        self.events = []
        self.dynamic_source_metadata = {}

    def copy(self, tag_offset=0):
        """
        Return a (shallow) copy of this object
        """
        new = self.__class__(self.origin)

        # Copy tags to given offset
        new.tags = {}
        for name, tag_list in self.tags.items():
            new.tags[name] = [tag.copy_modified(tag_offset) for tag in tag_list]

        # Creating copies of the events themselves should not be necessary
        new.events = self.events[:]
        new.dynamic_source_metadata = dict(self.dynamic_source_metadata)

        return new

    def __repr__(self):
        return repr(self.origin)

    def is_tracked(self):
        return bool(self.tags)

    def is_empty(self):
        return not self.is_tracked()

    def is_tagged(self, label):
        return label in self.tags

    def add_property(self, name, value):
        if not name or not value:
            return

        self.dynamic_source_metadata[name] = value

    def add_properties(self, dict_):
        if dict_:
            self.dynamic_source_metadata.update(dict_)

    def tag_keys(self):
        return self.tags.keys()

    def get_tags_at(self, index):
        """
        Find all of the tags that span a given index.
        """
        tags_at = set()

        for key, value in self.tags.items():
            for tag in value:
                if tag.covers(index):
                    tags_at.add(key)

        return list(tags_at)

    def last_tagged_index(self):
        """
        Returns the highest index of any tag range

        Since we can't reliably compute the length of the tagged object, we use this
        to determine the extent of all of the tag ranges that cover this object.
        """
        last_idx = 0

        for tag_list in self.tags.values():
            for tag in tag_list:
                if tag.end_index > last_idx:
                    last_idx = tag.end_index

        return last_idx

    def tags_at_range(self, tag_range):
        """
        Given a range, select all tags in that range the selected tags are
        shifted such that the start index of the new tag (0) aligns with the
        given start index in the range

        Example:
          current tags: 5-15
          range       : 5-10
          result      : 0-05
        :param tag_range: AdjustedSpan object with range of tag
        """
        location = {}

        length = tag_range.length()

        for key, value in self.tags.items():
            add = []

            for tag in value:
                comparison = tag.compare_range(tag_range.start, tag_range.stop)

                if comparison in [ABOVE, BELOW]:
                    continue
                elif comparison == LOW_SPAN:
                    add.append(Tag(tag.end_index - tag_range.start))
                elif comparison == HIGH_SPAN:
                    add.append(
                        Tag(
                            tag_range.stop - tag.start_index,
                            tag.start_index - tag_range.start,
                        )
                    )
                elif comparison == WITHOUT:
                    add.append(Tag(length))
                elif comparison == WITHIN:
                    add.append(Tag(tag.length, tag.start_index - tag_range.start))

            if len(add) != 0:
                location[key] = add

        return location

    def add_tag(self, label, tag_range):
        """
        Given a tag name and range object, add a new tag to this
        collection. If the given range touches an existing tag,
        we'll combine the two, adjusting the existing one and
        dropping this new one.

        :param label: tag label string
        :param tag_range: AdjustedSpan of tag range
        :return:
        """
        length = tag_range.length()
        tag = Tag(length, tag_range.start)
        self.add_existing_tag(label, tag)

    def add_existing_tag(self, label, tag):
        """
        Add an existing tag

        This function should really be called "add_tag", and the existing
        function with that name should be "add_tag_at".
        """
        existing = self.fetch_tags(label)
        self.set_tag(label, ordered_merge(existing, tag))

    def set_tag(self, label, tag_ranges):
        self.tags[label] = tag_ranges

    def delete_tag(self, label):
        try:
            del self.tags[label]
        except KeyError:
            pass

    def clear(self):
        self.tags = {}

    def cleanup_tags(self):
        """
        Merged tags that are touching or overlapping and delete empty tags
        """
        merge_tags(self.tags)

        # delete empty lists
        to_delete = [key for key, value in self.tags.items() if not value]
        for key in to_delete:
            del self.tags[key]

    def fetch_tags(self, label):
        return self.tags.get(label)

    ENCODED_MARKER = "_ENCODED"

    def fetch_encoded_tags(self):
        """
        Fetch all tags that end with *_ENCODED
        """
        return [
            self.tags[item] for item in self.tags if item.endswith(self.ENCODED_MARKER)
        ]

    def tags_to_ts_obj(self, splat_range=None):
        """
        Convert Tag objects to TraceTaintRange objects for teamserver
        """
        ranges = []
        for key, value in self.tags.items():
            if not value:
                continue

            for tag in value:
                taint_range = TraceTaintRange()

                taint_range.tag = key
                tag_range = splat_range or (tag.start_index, tag.end_index)
                taint_range.range = "{}:{}".format(*tag_range)

                ranges.append(taint_range)

        return ranges

    def delete_tags_at_ranges(self, ranges, shift=True):
        """
        Remove all tags within the given ranges.
        This does not delete an entire tag if part of that tag is
        outside this range, meaning we may reduce sizes of tags
        or split them.

        If shift is true, it is assumed the characters at those ranges were
        removed. If shift is false, it is assumed those ranges were replaced
        by the same number of characters and no shift is needed.

        current tags: 0-15
        range:        5-10
        result:       0-5, 10-15

        :param shift:
        :param ranges: list of AdjustedSpan objects
        """
        current_shift = 0

        for tag_range in ranges:
            self.remove_tag(tag_range)

            if shift:
                self.shift_tags_for_deletion(tag_range, current_shift)

                current_shift += tag_range.length()

        merge_tags(self.tags)

    def shift_tags(self, ranges):
        """
        Shift all the tags in this object by the given ranges.
        This method assumes the ranges are sorted, meaning
        the leftmost (lowest) range is first

        current tags: 0-15
        range:        5-10
        result:       0-5, 10-20
        """
        for item in ranges:
            self.shift_tags_for_insertion(item)

    def add_event(self, event):
        """
        Add event to properties
        """
        self.events.append(event)

        self.report_source()

        return self

    def build_event(
        self,
        policy_node,
        tagged,
        self_obj,
        ret,
        args,
        kwargs,
        parent_ids,
        possible_key=None,
        source_type=None,
        source_name=None,
    ):
        """
        Build ContrastEvent from an applied policy method

        After creation of the ContrastEvent, report the event to the observed_route in the RequestContext
        """
        event = ContrastEvent(
            policy_node,
            tagged,
            self_obj,
            ret,
            args,
            kwargs,
            parent_ids,
            possible_key,
            source_type,
            source_name,
        )
        self.add_event(event)

    def report_source(self):
        """
        This is for Route-based Auto-remediation

        We want to associate source events with routes. However, we only want to record
        sources if they are actually used. The reason for this is that the Python agent
        is fairly aggressive with source tracking, and so we preemptively track many
        sources even if they are never used. But to avoid inaccurate RBAV reports, we
        want to make sure sources are recorded only if they actually get used. For
        these purposes, being "used" means a source is the parent of another event.

        Criteria:
        - Previous event doesn't have a source type, do nothing
        - No current context and therefore no observed route, do nothing
        - There is an event of the same name and type, do nothing
        - Can create the TraceEventSource, append to the observed route sources

        :return: None
        """
        if len(self.events) < 2:
            return

        # Check whether the last event before the newest one was a source event.
        # Before this method is called, we add the newest event, which is now
        # self.events[-1]. So we check the event prior to that to see whether it's a
        # source event.
        event = self.events[-2]

        if not event.source_type:
            return

        context = contrast.CS__CONTEXT_TRACKER.current()

        if not context:
            return

        # ignore if the source already exists
        any_source_name_and_type_match = any(
            [
                source.type == event.source_type and source.name == event.source_name
                for source in context.observed_route.sources
            ]
        )

        if any_source_name_and_type_match:
            return

        source_event_obj = event.build_source_event(
            event.source_name if event.source_name is not None else ""
        )

        if not source_event_obj:
            return

        context.observed_route.sources.extend([source_event_obj])

    def remove_tag(self, tag_range):
        """
        Remove tag ranges in the given range
        :param tag_range: range of tags to delete
        :return: None
        """
        full_delete = []
        for key, tags in list(self.tags.items()):
            indexes_to_remove, tags_to_add = self._find_new_and_old_tags(
                tags, tag_range
            )

            for index in sorted(indexes_to_remove, reverse=True):
                del tags[index]

            ordered_merge(tags, tags_to_add)

            if not tags:
                full_delete.append(key)

        for item in full_delete:
            del self.tags[item]

    def _find_new_and_old_tags(self, tags, tag_range):
        indexes_to_remove = []
        tags_to_add = []

        for index, tag in enumerate(tags):
            comparison = tag.compare_range(tag_range.start, tag_range.stop)

            if comparison == LOW_SPAN:
                tag.set_end(tag_range.start)
            elif comparison == WITHIN:
                indexes_to_remove.append(index)
            elif comparison == WITHOUT:
                new_tag = copy(tag)
                new_tag.set_start(tag_range.stop)

                tags_to_add.append(new_tag)

                tag.set_end(tag_range.start)
            elif comparison == HIGH_SPAN:
                tag.set_start(tag_range.stop)
            elif comparison == ABOVE:
                continue

        return indexes_to_remove, tags_to_add

    def shift_tags_for_deletion(self, tag_range, shift):
        """
        Shift the tag ranges covering the given range
        We assume this is for a deletion, meaning we
        have to move tags to the left
        """
        for value in list(self.tags.values()):
            for tag in value:
                comparison = tag.compare_range(
                    tag_range.start - shift, tag_range.stop - shift
                )
                length = tag_range.length()

                if comparison == ABOVE:
                    tag.shift(0 - length)

    def shift_tags_for_insertion(self, tag_range):
        """
        Shift the tag ranges covering the given range
        We assume this is for a insertion, meaning we
        have to move tags to the right
        """
        for tags in list(self.tags.values()):
            add = []

            for tag in tags:
                comparison = tag.compare_range(tag_range.start, tag_range.stop)

                length = tag_range.length()

                if comparison in [LOW_SPAN, WITHOUT]:
                    add.append(
                        Properties.clone_tag_for_shift(tag, tag_range.start, length)
                    )
                elif comparison in [ABOVE, HIGH_SPAN, WITHIN]:
                    tag.shift(length)

            ordered_merge(tags, add)

    @staticmethod
    def clone_tag_for_shift(tag, new_start, length):
        """
        Copies a new tag and updates its start and end after shifting
        """
        new_tag = copy(tag)

        new_tag.set_start(new_start)

        new_tag.shift(length)

        tag.set_end(new_start)

        return new_tag



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.config.base_django_config_rule import (
    BaseDjangoConfigRule,
)
from .httponly_rule import HttpOnlyRuleMixin


class DjangoHttpOnlyRule(HttpOnlyRuleMixin, BaseDjangoConfigRule):
    SETTINGS_VALUE = "SESSION_COOKIE_HTTPONLY"



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from .splat_propagator import SplatPropagator


class CodecsSplatPropagator(SplatPropagator):
    def __init__(self, *args, **kwargs):
        """
        Most codecs methods return a tuple where the target string is the first element
        """
        super().__init__(*args, **kwargs)
        self.target = self.target[0]



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.agent_state import detected_deprecated_middleware
from contrast.asgi.middleware import ASGIMiddleware
from contrast.wsgi.middleware import WSGIMiddleware


def deprecated_wsgi(framework: str):
    class DeprecatedWSGIMiddleware(WSGIMiddleware):
        def __init__(self, *args, **kwargs):
            detected_deprecated_middleware(framework=framework, is_asgi=False)
            super().__init__(*args, **kwargs)

    return DeprecatedWSGIMiddleware


def deprecated_asgi(framework: str):
    class DeprecatedASGIMiddleware(ASGIMiddleware):
        def __init__(self, *args, **kwargs):
            detected_deprecated_middleware(framework=framework, is_asgi=True)
            super().__init__(*args, **kwargs)

    return DeprecatedASGIMiddleware



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys
import contrast
from contrast_vendor.wrapt import register_post_import_hook
from contrast.agent.policy import patch_manager
from contrast.assess_extensions import cs_str
from contrast.utils.patch_utils import build_and_apply_patch, wrap_and_watermark

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


START_METHOD = "start"
BOOTSTRAP_METHOD = "_bootstrap_inner"


def build_start_patch(orig_func, _):
    def start(wrapped, instance, args, kwargs):
        context = contrast.CS__CONTEXT_TRACKER.current()

        try:
            # Save the scope of the current active contextvars.Context to copy to the new thread
            instance.cs__parent_scope = cs_str.get_current_scope()
            instance.cs__parent_context = context
        except Exception:
            logger.exception("Failed to instrument thread start")

        return wrapped(*args, **kwargs)

    return wrap_and_watermark(orig_func, start)


def build_bootstrap_inner_patch(orig_func, _):
    def _bootstrap_inner(wrapped, instance, args, kwargs):  # pragma: no cover
        # The new thread inherits the scope from the thread that created it
        try:
            cs_str.set_exact_scope(instance.cs__parent_scope)
        except Exception:
            logger.exception("Failed to initialize thread scope")

        with contrast.CS__CONTEXT_TRACKER.lifespan(instance.cs__parent_context):
            # Ensure child thread still runs with the same parent request context
            # even if the parent thread has already exited as long as
            # the parent thread is in request context.
            result = wrapped(*args, **kwargs)

        # We expect result to be None, but this is done for consistency/safety
        return result

    return wrap_and_watermark(orig_func, _bootstrap_inner)


def patch_threading(threading_module):
    build_and_apply_patch(threading_module.Thread, START_METHOD, build_start_patch)
    # This instruments the method that actually runs inside the system thread
    build_and_apply_patch(
        threading_module.Thread, BOOTSTRAP_METHOD, build_bootstrap_inner_patch
    )


def register_patches():
    register_post_import_hook(patch_threading, "threading")


def reverse_patches():
    threading = sys.modules.get("threading")
    if not threading:
        return

    patch_manager.reverse_patches_by_owner(threading.Thread)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.protect.rule.base_rule import BaseRule


class UnsafeFileUpload(BaseRule):
    """
    Unsafe File Upload rule to protect against potentially malicious
    files that get uploaded
    """

    RULE_NAME = "unsafe-file-upload"



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import io

from contrast.agent.assess.utils import is_tracked
from contrast.utils.decorators import fail_quietly
from contrast.utils.assess.duck_utils import safe_getattr, safe_iterator

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


@fail_quietly("Failed check if obj is tracked", return_value=False)
def recursive_is_tracked(obj):
    """
    This is a fairly hot code path in some applications; be mindful of performance
    """
    if obj is None:
        return False

    if isinstance(obj, dict):
        for key, value in obj.items():
            if recursive_is_tracked(key) or recursive_is_tracked(value):
                return True
    elif isinstance(obj, io.IOBase):
        return safe_getattr(obj, "cs__tracked", False) or safe_getattr(
            obj, "cs__source", False
        )
    # These are the only safe objects where we can guarantee that we won't
    # destructively modify the object such that the application may see
    # different results.
    elif isinstance(obj, (list, tuple)):
        for item in safe_iterator(obj):
            if recursive_is_tracked(item):
                return True
        return False

    return is_tracked(obj)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys
from contrast_vendor.wrapt import register_post_import_hook

from contrast.agent import scope
from contrast.agent.policy import patch_manager
from contrast.agent.assess.policy import propagation_policy
from contrast.agent.assess.policy.source_policy import apply_stream_source
from contrast.agent.assess.policy.preshift import Preshift
from contrast.agent.assess.policy.propagators import STREAM_ACTIONS, stream_propagator
from contrast.agent.assess.utils import get_properties
from contrast.utils.patch_utils import wrap_and_watermark
from contrast.assess_extensions import cs_str

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


def build_and_patch_read_method(io_type, method_name):
    orig_func = getattr(io_type, method_name)
    if patch_manager.is_patched(orig_func):
        return

    propagation_method = STREAM_ACTIONS.get(method_name)

    def patched_func(wrapped, instance, args, kwargs):
        """
        This function has been optimized for performance. It looks similar to others in
        this module; however, do not attempt to deduplicate unless you're sure of the
        performance impact.
        """
        result = wrapped(*args, **kwargs)

        if cs_str.in_contrast_or_propagation_scope():
            return result

        scope.enter_propagation_scope()

        try:
            props = get_properties(instance)
            if props is not None and props.tags:
                preshift = Preshift(instance, args, kwargs)
                propagation_method(
                    method_name,
                    preshift,
                    result,
                    result,
                )
            elif instance.cs__source:
                apply_stream_source(method_name, result, instance, result, args, kwargs)
        except Exception as ex:
            _debug_log(method_name, ex)

        scope.exit_propagation_scope()

        return result

    patch = wrap_and_watermark(orig_func, patched_func)
    patch_manager.patch(io_type, method_name, patch)


def build_and_patch_write_method(io_type, method_name):
    orig_func = getattr(io_type, method_name)
    if patch_manager.is_patched(orig_func):
        return

    def patched_func(wrapped, instance, args, kwargs):
        """
        This function has been optimized for performance. It looks similar to others in
        this module; however, do not attempt to deduplicate unless you're sure of the
        performance impact.
        """
        result = wrapped(*args, **kwargs)

        if cs_str.in_contrast_or_propagation_scope():
            return result

        scope.enter_propagation_scope()

        try:
            preshift = Preshift(instance, args, kwargs)
            stream_propagator.propagate_stream_write(
                method_name, preshift, instance, result
            )
        except Exception as ex:
            _debug_log(method_name, ex)

        scope.exit_propagation_scope()

        return result

    patch = wrap_and_watermark(orig_func, patched_func)
    patch_manager.patch(io_type, method_name, patch)


def build_and_patch_writelines(io_type, method_name):
    orig_func = getattr(io_type, method_name)
    if patch_manager.is_patched(orig_func):
        return

    def patched_func(wrapped, instance, args, kwargs):
        """
        This function has been optimized for performance. It looks similar to others in
        this module; however, do not attempt to deduplicate unless you're sure of the
        performance impact.
        """
        args_list = list(args[0:1]) + list(args[1:])

        result = wrapped(*args_list, **kwargs)

        if cs_str.in_contrast_or_propagation_scope():
            return result

        scope.enter_propagation_scope()

        try:
            propagation_policy.propagate_stream(
                method_name,
                result,
                instance,
                result,
                args_list,
                kwargs,
            )
        except Exception as ex:
            _debug_log(method_name, ex)

        scope.exit_propagation_scope()

        return result

    patch = wrap_and_watermark(orig_func, patched_func)
    patch_manager.patch(io_type, method_name, patch)


def _debug_log(method_name, ex):
    logger.debug("failed to propagate %s", method_name, exc_info=ex)


def patch_getvalue(io_module):
    build_and_patch_read_method(io_module.StringIO, "getvalue")
    build_and_patch_read_method(io_module.BytesIO, "getvalue")


def patch_io(io_module):
    """
    Apply patches to methods of builtin stream types
    """
    for io_type in [io_module.StringIO, io_module.BytesIO]:
        build_and_patch_write_method(io_type, "write")

    # No need to hook StringIO.writelines because it is implemented as str.join under
    # the hood, so we already propagate for free. Unfortunately this might make the
    # reporting look a little odd, so we maybe should consider another solution later.
    build_and_patch_writelines(io_module.BytesIO, "writelines")

    # This patch exists solely for the purposes of working around our inability
    # to patch getvalue in StringIO and BytesIO using funchook.
    patch_getvalue(io_module)


def register_patches():
    register_post_import_hook(patch_io, "io")


def reverse_patches():
    io_module = sys.modules.get("io")
    if not io_module:
        return

    patch_manager.reverse_patches_by_owner(io_module)
    patch_manager.reverse_patches_by_owner(io_module.BytesIO)
    patch_manager.reverse_patches_by_owner(io_module.StringIO)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import logging
from sys import version
from sysconfig import get_platform

import contrast
from contrast.agent.assess.utils import get_properties, track_and_tag
from contrast.agent.patch_controller import enable_assess_patches
from contrast.agent.settings import Settings
from contrast.assess_extensions import cs_str
from contrast.utils.loggers.logger import setup_basic_agent_logger


class ContextDuck:
    @property
    def stop_propagation(self):
        return False

    def propagated(self):
        """No action required, just a dummy method to satisfy the expected API"""
        pass


class MyString(str):
    pass


def test_concat():
    x = "foo"
    y = "bar"
    track_and_tag(x)
    return x + y


def test_join():
    x = "sep"
    track_and_tag(x)
    return x.join(["foo", "bar"])


def test_cformat():
    x = "whatever"
    y = "%s in here"
    track_and_tag(x)
    return y % x  # pylint: disable=string-format-interpolation


def test_repr():
    x = b"whatever"
    track_and_tag(x)
    return repr(x)


def test_string_subclass():
    x = "whatever"
    track_and_tag(x)
    return MyString(x)


def test_string_subclass_defined_later():
    x = "whatever"
    track_and_tag(x)

    class MyString(str):
        pass

    return MyString(x)


def test_django_safestring():
    from django.utils.safestring import SafeString

    x = "whatever"
    track_and_tag(x)
    return str(SafeString(x))


PROPAGATOR_TESTS = {
    "cformat": test_cformat,
    "concat": test_concat,
    "join": test_join,
    "repr": test_repr,
    "string subclass": test_string_subclass,
    "string subclass (defined after hook)": test_string_subclass_defined_later,
    "safestring": test_django_safestring,
}


def propagator_check():
    # We don't necessarily need logs from this utility
    setup_basic_agent_logger(level=logging.ERROR)

    settings = Settings()
    settings.config.put("assess.enable", True)

    # NOTE: it might be nice to add an option to disable extension patches
    cs_str.init_contrast_scope_cvars()
    enable_assess_patches()

    # NOTE: here's where we might eventually apply the rewriter

    print("Contrast Agent Propagator Diagnostic Tool")
    print("Platform: {}, Python: {}\n".format(get_platform(), version))

    num_fails = 0
    with contrast.CS__CONTEXT_TRACKER.lifespan(ContextDuck()):
        for name, tester in PROPAGATOR_TESTS.items():
            is_ok = get_properties(tester()) is not None
            # NOTE: maybe add command-line option for no unicode/emoji output
            print("✅" if is_ok else "❌", name)
            if not is_ok:
                num_fails += 1
            contrast.STRING_TRACKER.clear()

    total = len(PROPAGATOR_TESTS)
    num_passes = total - num_fails
    print("{} passed, {} failed of {} total".format(num_passes, num_fails, total))

    return int(bool(num_fails))



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import re

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class MetricsDict(dict):
    def __init__(self, value_type):
        self._value_type = value_type

    def __setitem__(self, key, value):
        if not isinstance(key, str):
            logger.debug(
                "WARNING: non-string key will be omitted from telemetry metrics",
                key=key,
                value=value,
            )
            return
        if not re.fullmatch(r"[a-zA-Z0-9\._-]{1,63}", key):
            logger.debug(
                "WARNING: invalid key will be omitted from telemetry metrics",
                key=key,
                value=value,
            )
            return
        if len(key) >= 28:
            # we enforce this condition separately from the regex in order to exactly
            # align with regex in the spec
            logger.debug(
                "WARNING: too-long key will be omitted from telemetry metrics",
                key=key,
                value=value,
            )
            return
        if not isinstance(value, self._value_type):
            logger.debug(
                "WARNING: wrong-type value will be omitted from telemetry metrics",
                key=key,
                value=value,
                expected_type=self._value_type,
            )
            return
        if self._value_type == str and len(value) == 0:
            logger.debug(
                "WARNING: blank string value will be omitted from telemetry metrics",
                key=key,
                value=value,
                expected_type=self._value_type,
            )
            return
        if self._value_type == str and len(value) > 200:
            logger.debug(
                "WARNING: too-long string value will be omitted from telemetry metrics",
                key=key,
                value=value,
            )
            return

        key = key.lower()
        super().__setitem__(key, value)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import re

from contrast.agent.assess.contrast_event import ContrastEvent
from contrast.agent.assess.properties import Properties
from contrast.agent.assess.utils import get_properties
from contrast.utils.assess.duck_utils import is_iterable, safe_getattr, safe_iterator


def cs__apply_trigger(
    context, rule, node, orig_source, self_obj, ret, possible_key, args, kwargs
):
    if not context or not node:
        return

    if rule.disabled:
        return

    source = rule.extract_source(node, orig_source, args, kwargs)

    if not node.dataflow_rule:
        if node.good_value:
            if not isinstance(source, str) or re.match(node.good_value, source):
                return
        elif node.bad_value:
            if not isinstance(source, str) or not re.match(node.bad_value, source):
                return

        build_finding(
            context, rule, node, orig_source, self_obj, ret, None, args, kwargs
        )

        return

    if rule.is_violated(node, source, **dict(orig_args=args)):
        build_finding(
            context, rule, node, orig_source, self_obj, ret, possible_key, args, kwargs
        )

    elif isinstance(source, dict):
        for key, value in source.items():
            cs__apply_trigger(
                context, rule, node, key, self_obj, ret, None, args, kwargs
            )
            cs__apply_trigger(
                context, rule, node, value, self_obj, ret, key, args, kwargs
            )

    elif is_iterable(source):
        for value in safe_iterator(source):
            cs__apply_trigger(
                context, rule, node, value, self_obj, ret, None, args, kwargs
            )


def build_finding(
    context,
    rule,
    node,
    target,
    self_obj,
    ret,
    possible_key,
    args,
    kwargs,
    target_properties=None,
):
    """
    Builds a finding and appends it to the current context

    :param context: current RequestContext
    :param rule: TriggerRule that was violated
    :param node: Weaver that was used to identify the source
    :param target: tracked item that triggered the rule
    :param self_obj: object of the the called node method ; could be None if it was a module-level function
    :param ret: return of the policy method
    :param possible_key: possible key of the value in the kwarg
    :param args: tuple of methods arguments
    :param kwargs: dictionary of methods keyword arguments
    :return: None
    """
    events = []
    properties = {}

    if target is not None and target_properties is None:
        # If the target is a stream that is being treated as a source, then we
        # build a new Properties object for it and add the stream's source
        # event to it. In this case the stream object itself is actually
        # triggering the rule, which means that no data was necessarily read
        # from the stream before triggering the rule.
        if safe_getattr(target, "cs__source", False):
            target_properties = Properties(target)
            target_properties.add_event(target.cs__source_event)
        else:
            target_properties = get_properties(target)

    if target_properties is not None:
        events = [event.to_reportable_event() for event in target_properties.events]

        target_props = target_properties.dynamic_source_metadata
        if target_props:
            for key, value in target_props.items():
                properties[key] = value

    parent_ids = (
        [target_properties.events[-1].event_id]
        if target_properties and target_properties.events
        else []
    )

    contrast_event = ContrastEvent(
        node,
        target,
        self_obj,
        ret,
        args,
        kwargs,
        parent_ids,
        possible_key,
    ).to_reportable_event()

    events.append(contrast_event)

    rule.build_and_append_finding(
        context, properties, node, target, events=events, source=target
    )



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import requests

from .base_ts_message import BaseTsAppMessage
from contrast.reporting.activity_masker import ActivityMasker
from contrast.utils.decorators import fail_loudly
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class ApplicationActivity(BaseTsAppMessage):
    def __init__(self, context):
        super().__init__()
        ActivityMasker(context).mask_sensitive_data()

        self.body = {"lastUpdate": self.since_last_update}

        self.body["inventory"] = {
            # Used by TeamServer to aggregate counts across a given time period, for
            # Protect and attacker activity.
            "activityDuration": 0,
            "components": [],
        }
        if context.request and context.request.user_agent:
            self.body["inventory"]["browsers"] = [context.request.user_agent]
        if context.database_info:
            for db_info in context.database_info:
                self.body["inventory"]["components"].append(db_info)

        if context.attacks:
            self.body["defend"] = {"attackers": []}

            for attack in context.attacks:
                self.body["defend"]["attackers"].append(
                    {
                        "protectionRules": {
                            attack.rule_id: attack.to_json(context.request)
                        },
                        "source": {
                            "ip": context.request.client_addr or "",
                            "xForwardedFor": context.request.headers.get(
                                "X-Forwarded-For"
                            )
                            or "",
                        },
                    }
                )

    @property
    def name(self):
        return "application-activity"

    @property
    def path(self):
        return "activity/application"

    @property
    def request_method(self):
        return requests.put

    @property
    def expected_response_codes(self):
        return [200, 204]

    @fail_loudly("Failed to process Activity response")
    def process_response(self, response, reporting_client):
        self.process_response_code(response, reporting_client)



# SPDX-License-Identifier: MIT OR Apache-2.0
# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

"""
Generic utilities.
"""

from __future__ import annotations

import errno
import sys

from contextlib import suppress
from typing import Any, Callable


def until_not_interrupted(f: Callable[..., Any], *args: Any, **kw: Any) -> Any:
    """
    Retry until *f* succeeds or an exception that isn't caused by EINTR occurs.

    :param f: A callable like a function.
    :param *args: Positional arguments for *f*.
    :param **kw: Keyword arguments for *f*.
    """
    while True:
        try:
            return f(*args, **kw)
        except OSError as e:
            if e.args[0] == errno.EINTR:
                continue
            raise


def get_processname() -> str:
    # based on code from
    # https://github.com/python/cpython/blob/313f92a57bc3887026ec16adb536bb2b7580ce47/Lib/logging/__init__.py#L342-L352
    processname = "n/a"
    mp: Any = sys.modules.get("multiprocessing")
    if mp is not None:
        # Errors may occur if multiprocessing has not finished loading
        # yet - e.g. if a custom import hook causes third-party code
        # to run when multiprocessing calls import.
        with suppress(Exception):
            processname = mp.current_process().name

    return processname



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent import scope
from contrast.agent.assess.policy.analysis import skip_analysis
from contrast.agent.assess.policy.source_node import SourceNode
from contrast.agent.assess.policy.source_policy import cs__apply_source
from contrast.utils.assess.stream_utils import ContrastWsgiStreamProxy
from contrast.utils.decorators import fail_loudly

# a best effort at translating the types in this document:
# https://bitbucket.org/contrastsecurity/teamserver/src/
# 862d499227e8eac42b6eb7c6b03f10b7f1556218/teamserver-agent-messages/src/main/java/
# contrast/agent/messages/finding/trace/EventSourceTypeDTM.java#lines-8
ENVIRON_SOURCES = {
    "CONTENT_LENGTH": "HEADER",
    "CONTENT_TYPE": "HEADER",
    "PATH_INFO": "URI",
    "QUERY_STRING": "QUERYSTRING",
    "REMOTE_ADDR": "URI",
    "REMOTE_HOST": "URI",
    "REMOTE_USER": "URI",
    "SCRIPT_NAME": "HEADER",
    "SERVER_NAME": "HEADER",
    "SERVER_PORT": "HEADER",
    "SERVER_PROTOCOL": "HEADER",
    "webob.adhoc_attrs": "OTHER",
    "webob.url_encoding": "OTHER",
}

SOURCE_TAGS = {"HEADER": ["NO_NEWLINES"], "PARAMETER": ["CROSS_SITE"]}

SOURCE_DICT = {"module": "wsgi.environ", "instance_method": False, "target": "RETURN"}

WEBOB_KEYS_TO_DELETE = {
    "webob._parsed_query_vars",
    "webob._parsed_post_vars",
    "webob._body_file",
    "webob._parsed_cookies",
    "webob._cache_control",
}


@fail_loudly("failed to track cookie sources")
def track_cookie_sources(framework, context, cookies):
    """track cookie keys and values"""
    if skip_analysis(context):
        return

    with scope.contrast_scope():
        _track_keys_and_values(
            framework, context, cookies, "COOKIE_KEY", "COOKIE", no_cross_site=True
        )


@fail_loudly("Failed to track environ sources")
def track_environ_sources(framework, context, environ):
    """
    This method will track necessary information in the environ

    For wsgi frameworks, this is the one true source of (stateless) untrusted data

    :param framework: current application's framework
    :param context: current request context
    :param environ: WSGI environ dict

    :return: None
    """
    if skip_analysis(context):
        return

    with scope.contrast_scope():
        for key, value in environ.items():
            _track_environ_item(framework, context, environ, key, value)

        _make_request_body_seekable(context.request)
        _track_wsgi_input(environ)
        _remove_webob_environ_vars(environ)


def _build_source_node(framework, method_name, source_type, no_cross_site=False):
    """
    Builds a new SourceNode based on the method name and source type

    Since cookie sources are not tagged with CROSS_SITE, callers must indicate whether
    the given source corresponds to a cookie or not.

    :param framework: str framework name for node
    :param method_name: name of the environ key
    :param source_type: type of the source
    :param no_cross_site: boolean indicating whether source tags include CROSS_SITE
    :return: SourceNode for cs__apply_source
    """

    source_dict = SOURCE_DICT.copy()

    tags = [] if no_cross_site else SOURCE_TAGS.get(source_type, ["CROSS_SITE"])

    node = SourceNode(
        method_name=method_name,
        node_type=source_type,
        tags=tags,
        **source_dict,
    )
    node.skip_stacktrace = True

    return node


def _track_keys_and_values(
    framework, context, mapping, key_type, value_type, no_cross_site=False
):
    for key, value in mapping.items():
        if key:
            key_node = _build_source_node(framework, key, key_type, no_cross_site)
            cs__apply_source(
                context, key_node, key, mapping, key, (), {}, source_name=key
            )

        if value:
            value_node = _build_source_node(framework, value, value_type, no_cross_site)
            cs__apply_source(
                context, value_node, value, mapping, value, (), {}, source_name=key
            )


def _track_environ_item(framework, context, environ, key, value):
    source_type = ""
    if key in ENVIRON_SOURCES:
        source_type = ENVIRON_SOURCES[key]

        node = _build_source_node(framework, key, source_type, False)

        cs__apply_source(context, node, value, environ, value, (), {}, source_name=key)
    elif key.startswith("HTTP_"):
        # This is to track custom headers we may not know about
        source_type = "COOKIE" if key == "HTTP_COOKIE" else "HEADER"

        node = _build_source_node(framework, key, source_type, "COOKIE" in key)

        cs__apply_source(context, node, value, environ, value, (), {}, source_name=key)

    # Track HTTP header keys as well
    if key.startswith("HTTP_") or source_type == "HEADER":
        no_cross_site = (source_type == "HEADER") or (key == "HTTP_COOKIE")

        key_node = _build_source_node(framework, key, "HEADER_KEY", no_cross_site)
        cs__apply_source(context, key_node, key, environ, key, (), {}, source_name=key)


def _make_request_body_seekable(request):
    """
    Before we track wsgi.input, we ask our vendored webob to manipulate the input stream
    to make it seekable. There is no security risk here, because this simply involves
    reading raw bytes from wsgi.input into a seekable stream.

    We need to do this in case the application or any downstream middleware uses webob:
    1. Correctness: we would lose propagation from the original wsgi.input to the new
       stream if make_body_seekable is called for the first time after environ tracking.
    2. Performance: there's no need to propagate through make_body_seekable since we can
       just make the resulting seekable stream a source.
    Webob knows if this method has ever been called by storing the
    `webob.is_body_seekable` boolean directly in the environ dict.

    There is also no risk to correctness despite the fact that `make_body_seekable`
    resets the stream to position 0. This is because middlewares should not consume this
    stream without resetting it - so if we receive a partially-consumed stream, an
    earlier middleware is at fault.

    We should only track wsgi.input after calling this method.

    :param request: contrast.agent.request.Request instance
    :return: None
    """
    request.make_body_seekable()


def _track_wsgi_input(environ):
    wsgi_input = "wsgi.input"

    if wsgi_input in environ:
        if not hasattr(environ[wsgi_input], "cs__source"):
            environ[wsgi_input] = ContrastWsgiStreamProxy(environ[wsgi_input])

        environ[wsgi_input].cs__source = True
        environ[wsgi_input].cs__source_type = _get_source_by_content_type(environ)
        environ[wsgi_input].cs__source_tags = ["UNTRUSTED", "CROSS_SITE"]


def _get_source_by_content_type(environ):
    """
    Define the wsgi.input source type based on the request content type.
    If no specific content-type, the fault soruce type is BODY.
    """
    source_type = "BODY"

    content_type = environ.get("CONTENT_TYPE", "")
    if content_type == "application/x-www-form-urlencoded":
        source_type = "PARAMETER"
    elif content_type.startswith("multipart/"):
        source_type = (
            "MULTIPART_FORM_DATA"
            if "form-data" in content_type
            else "MULTIPART_CONTENT_DATA"
        )
    return source_type


@fail_loudly("Failed to remove webob environ variables")
def _remove_webob_environ_vars(environ):
    """
    webob adds several variables to the environ for efficiency. For example, if webob
    ever parses wsgi.input to extract request body parameters, it stores them in a
    MultiDict under environ['webob._parsed_post_vars']. This way, if a future middleware
    or application using webob tries to access the post params again, it doesn't need to
    re-parse the request body.

    This is a good idea, but it's really a pain for us. It means that it's not enough
    to simply track wsgi.input, because if it has already been parsed once by the time
    we call the environ tracker, data stored under these special webob.* keys won't be
    tracked. Here, we delete these variables, forcing webob to re-parse any necessary
    info using our now-tracked environ values. In the future, we should instead iterate
    over these entries and track their values explicitly, for the sake of performance.

    Another consequence of deleting these variables is that their values will
    necessarily be webob types for which we have patches. Since our request wrapper is a
    subclass of contrast_vendor.webob.BaseRequest, we were ending up with _our_ vendored
    webob objects in the application's environ. We don't have or want patches for our
    own vendored webob, so this ensures that all of webob's cached environ values will
    have patches.

    We should always check to see if these values are present, even if we're inside
    of a framework that doesn't explicitly use webob. This is because it's possible that
    an earlier middleware used webob, causing the values to be added to environ.

    To the best of our current understanding, the following extra environ items are
    of interest:

    webob._parsed_query_vars
        A 2-tuple
        - first element is a MultiDict of parsed GET parameters
        - second element is the query string that was parsed

    webob._parsed_post_vars
        A 2-tuple
        - the first element is a MultiDict of parsed POST parameters
        - the second is the body stream that was parsed

    webob.adhoc_attrs
        A custom dictionary of arbitrary data passed between middlewares. We make a best
        effort to patch this, and we don't delete it, because we can't know much about
        it.

    webob._body_file
        A 2-tuple
        - first element is a double-wrapper for the raw wsgi.input stream
            BufferedReader(webob.LimitedLengthFile(wsgi.input))
        - second element is the underlying raw wsgi.input stream
            It should be PEP-333 compliant

    webob.is_body_seekable
        A boolean
        - Created/set to True once webob replaces wsgi.input with a seekable file-like
          object
        - Assumed to be False if missing
        - We do not remove this from environ, because we don't want webob to replace
          wsgi.input if possible (since by this point we've made it a source)

    webob._parsed_cookies
        A 2-tuple
        - first element is a dict of cookie keys to cookie values (parsed from
          HTTP_COOKIE)
        - second element is the HTTP_COOKIE string that was parsed to produce element #1

    webob._cache_control
        A 2-tuple
        - first element is the `HTTP_CACHE_CONTROL` header from the WSGI environ
        - second element is webob's parsed representation, a CacheControl object

    webob.url_encoding
        A string
        - Used to store the string encoding of the URL. We don't delete it because it's
          easy to track directly. In the future we might want to look into whether we
          need to track this at all.

    :param environ: WSGI environ dict
    :return: None
    """
    for key in WEBOB_KEYS_TO_DELETE:
        if key in environ:
            del environ[key]



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
class CustomSearcher(object):
    """
    Base object for Custom Searchers that rules can implement
    """

    IMPACT_NONE = 0
    IMPACT_LOW = 1
    IMPACT_MEDIUM = 2
    IMPACT_HIGH = 3
    IMPACT_CRITICAL = 4

    def __init__(self, searcher_id):
        self._searcher_id = searcher_id

    @property
    def searcher_id(self):
        return self._searcher_id

    def impact_of(self, value):
        return NotImplementedError



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from .fix_interpreter_permissions import fix_interpreter_permissions  # noqa
from .propagator_check import propagator_check  # noqa
from .validate_config import validate_config  # noqa
from .runner import runner  # noqa



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import contrast
from contrast.agent.assess.adjusted_span import AdjustedSpan
from contrast.utils.assess.duck_utils import safe_getattr
from contrast.utils.assess.tag_utils import merge_tags


def get_properties(value):
    return safe_getattr(value, "cs__properties", None)


def set_properties(obj, props):
    obj.cs__properties = props


def clear_properties():
    contrast.STRING_TRACKER.clear()


def is_trackable(obj):
    return hasattr(obj, "cs__properties")


def is_tracked(value):
    return bool(get_properties(value))


def track_string(value):
    return contrast.STRING_TRACKER.track(value)


def track_and_tag(value, label="UNTRUSTED", start=None, end=None):
    properties = track_string(value)
    if not properties:
        return None

    if start is None:
        start = 0
    if end is None:
        end = len(value)

    properties.add_tag(label, AdjustedSpan(start, end))
    return properties


def copy_events(target_props, source_props):
    if source_props is None or target_props is None or target_props is source_props:
        return

    for event in source_props.events:
        target_props.events.append(event)


def copy_from(to_obj, from_obj, shift=0, skip_tags=None):
    """Copy events and tags from from_obj to to_obj"""
    if from_obj is to_obj:
        return

    if not is_tracked(from_obj):
        return

    from_props = get_properties(from_obj)

    # we assume to_obj has already been tracked and has properties
    to_props = get_properties(to_obj)

    if from_props == to_props:
        return

    copy_events(to_props, from_props)

    for key in from_props.tag_keys():
        if skip_tags and key in skip_tags:
            continue

        new_tags = []

        from_props_tags = from_props.fetch_tags(key)

        for tag in from_props_tags:
            new_tags.append(tag.copy_modified(shift))

        existing_tags = to_props.fetch_tags(key)

        if existing_tags:
            existing_tags.extend(new_tags)
        else:
            to_props.set_tag(key, new_tags)


def get_self_for_method(patch_policy, args):
    """
    Retrieves self for a method's PatchLocationPolicy,

    If any node in the policy has a False instance_method attribute return None
    """
    for node in patch_policy.all_nodes:
        if not node.instance_method:
            return None

    return args[0] if args else None


def get_last_event_id(source_properties):
    if source_properties.events and source_properties.events[-1]:
        last_event = source_properties.events[-1]
        return last_event.event_id
    return None


def get_last_event_ids_from_sources(sources):
    """
    Gathers from given sources the parent IDs that should be used for an event
    """
    id_list = []

    for source in sources:
        source_properties = get_properties(source)
        if source_properties is None:
            continue

        event_id = get_last_event_id(source_properties)
        if event_id is not None and event_id not in id_list:
            id_list.append(event_id)

    return id_list


def copy_tags_in_span(target, source_properties, span, offset=0):
    """
    Given source properties, copies tags at a given span to the target
    """
    span = AdjustedSpan(*tuple(x + offset for x in span))
    source_tags = source_properties.tags_at_range(span)
    if not source_tags:
        return get_properties(target)

    target_properties = track_string(target)
    if target_properties is None:
        return get_properties(target)

    for name, tags in source_tags.items():
        for tag in tags:
            target_properties.add_existing_tag(name, tag)

    merge_tags(target_properties.tags)
    return target_properties


def copy_tags_to_offset(target_properties, source_tags, target_offset):
    """
    Given source tags, copy to target properties at offset.

    The caller is responsible for updating the string tracker if necessary.
    """
    for name, tags in source_tags.items():
        for tag in tags:
            new_tag = tag.copy_modified(target_offset)
            target_properties.add_existing_tag(name, new_tag)
